# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers for flags in commands working with Google Cloud Functions."""


import json

from argcomplete.completers import DirectoriesCompleter
from googlecloudsdk.api_lib.functions.v1 import util as api_util
from googlecloudsdk.api_lib.functions.v2 import client as client_v2
from googlecloudsdk.calliope import actions
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.calliope.concepts import concepts
from googlecloudsdk.calliope.concepts import deps
from googlecloudsdk.command_lib.eventarc import flags as eventarc_flags
from googlecloudsdk.command_lib.util import completers
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core import resources
from googlecloudsdk.core.console import console_io
import six

API = 'cloudfunctions'
API_VERSION = 'v1'
LOCATIONS_COLLECTION = API + '.projects.locations'

SIGNATURE_TYPES = ['http', 'event', 'cloudevent']
SEVERITIES = ['DEBUG', 'INFO', 'ERROR']
EGRESS_SETTINGS = ['PRIVATE-RANGES-ONLY', 'ALL']
INGRESS_SETTINGS = ['ALL', 'INTERNAL-ONLY', 'INTERNAL-AND-GCLB']
SECURITY_LEVEL = ['SECURE-ALWAYS', 'SECURE-OPTIONAL']
INGRESS_SETTINGS_MAPPING = {
    'ALLOW_ALL': 'all',
    'ALLOW_INTERNAL_ONLY': 'internal-only',
    'ALLOW_INTERNAL_AND_GCLB': 'internal-and-gclb',
}

EGRESS_SETTINGS_MAPPING = {
    'PRIVATE_RANGES_ONLY': 'private-ranges-only',
    'ALL_TRAFFIC': 'all',
}
DIRECT_VPC_EGRESS_SETTINGS_MAPPING = {
    'VPC_EGRESS_PRIVATE_RANGES_ONLY': 'private-ranges-only',
    'VPC_EGRESS_ALL_TRAFFIC': 'all',
}

SECURITY_LEVEL_MAPPING = {
    'SECURE_ALWAYS': 'secure-always',
    'SECURE_OPTIONAL': 'secure-optional',
}

_KMS_KEY_NAME_PATTERN = (
    r'^projects/[^/]+/locations/[^/]+/keyRings/[a-zA-Z0-9_-]+'
    '/cryptoKeys/[a-zA-Z0-9_-]+$'
)
_KMS_KEY_NAME_ERROR = (
    'KMS key name should match projects/{project}/locations/{location}'
    '/keyRings/{keyring}/cryptoKeys/{cryptokey} and only contain characters '
    'from the valid character set for a KMS key.'
)
_DOCKER_REPOSITORY_NAME_RESOURCE_PATTERN = (
    r'^projects/[^/]+/locations/[^/]+/repositories/[a-z]([a-z0-9-]*[a-z0-9])?$'
)
_DOCKER_REPOSITORY_NAME_PKG_PATTERN = (
    r'^(?P<location>.*)-docker.pkg.dev\/(?P<project>[^\/]+)\/(?P<repo>[^\/]+)'
)
_DOCKER_REPOSITORY_NAME_PATTERN = '({}|{})'.format(
    _DOCKER_REPOSITORY_NAME_RESOURCE_PATTERN,
    _DOCKER_REPOSITORY_NAME_PKG_PATTERN,
)
_DOCKER_REPOSITORY_NAME_ERROR = (
    'Docker repository name should match'
    ' `projects/{project}/locations/{location}/repositories/{repository}` or'
    ' `{location}-docker.pkg.dev/{project}/{repository}` and only contain'
    ' characters from the valid character set for a repository.'
)

DOCKER_REGISTRY_MAPPING = {
    'CONTAINER_REGISTRY': 'container-registry',
    'ARTIFACT_REGISTRY': 'artifact-registry',
}

RUNTIME_UPDATE_POLICY_MAPPING = {
    'AUTOMATIC': 'automatic',
    'ON_DEPLOY': 'on-deploy',
}


def AddMinLogLevelFlag(parser):
  min_log_arg = base.ChoiceArgument(
      '--min-log-level',
      choices=[x.lower() for x in SEVERITIES],
      help_str='Minimum level of logs to be fetched.',
  )
  min_log_arg.AddToParser(parser)


def AddIngressSettingsFlag(parser):
  ingress_settings_arg = base.ChoiceArgument(
      '--ingress-settings',
      choices=[x.lower() for x in INGRESS_SETTINGS],
      help_str=(
          'Ingress settings controls what traffic can reach the '
          'function. By default `all` will be used.'
      ),
  )
  ingress_settings_arg.AddToParser(parser)


def AddEgressSettingsFlag(parser):
  egress_settings_arg = base.ChoiceArgument(
      '--egress-settings',
      choices=[x.lower() for x in EGRESS_SETTINGS],
      help_str=(
          'Egress settings controls what traffic is diverted through the '
          'VPC Access Connector resource. '
          'By default `private-ranges-only` will be used.'
      ),
  )
  egress_settings_arg.AddToParser(parser)


def AddSecurityLevelFlag(parser):
  security_level_arg = base.ChoiceArgument(
      '--security-level',
      default='secure-always',
      choices=[x.lower() for x in SECURITY_LEVEL],
      help_str=(
          "Security level controls whether a function's URL supports "
          'HTTPS only or both HTTP and HTTPS. By default, `secure-always` will'
          ' be used, meaning only HTTPS is supported.'
      ),
  )
  security_level_arg.AddToParser(parser)


def GetLocationsUri(resource):
  registry = resources.REGISTRY.Clone()
  registry.RegisterApiByName(API, API_VERSION)
  ref = registry.Parse(
      resource.name,
      params={'projectsId': properties.VALUES.core.project.GetOrFail},
      collection=LOCATIONS_COLLECTION,
  )
  return ref.SelfLink()


def AddFunctionMemoryAndCpuFlags(parser):
  """Add flags for specifying function memory and cpu to the parser."""

  memory_help_text = """\
  Limit on the amount of memory the function can use.

  Allowed values for v1 are: 128MB, 256MB, 512MB, 1024MB, 2048MB, 4096MB,
  and 8192MB.

  Allowed values for GCF 2nd gen are in the format: <number><unit> with allowed units
  of "k", "M", "G", "Ki", "Mi", "Gi". Ending 'b' or 'B' is allowed, but both are
  interpreted as bytes as opposed to bits.

  Examples: 1000000K, 1000000Ki, 256Mb, 512M, 1024Mi, 2G, 4Gi.

  By default, a new function is limited to 256MB of memory. When
  deploying an update to an existing function, the function keeps its old
  memory limit unless you specify this flag."""
  group = parser.add_group(required=False)
  cpu_help_text = """\
    The number of available CPUs to set. Only valid when
    `--memory=MEMORY` is specified.

    Examples: .5, 2, 2.0, 2000m.

    By default, a new function's available CPUs is determined based on its memory value.

    When deploying an update that includes memory changes to an existing function,
    the function's available CPUs will be recalculated based on the new memory unless this flag
    is specified. When deploying an update that does not include memory changes to an existing function,
    the function's "available CPUs" setting will keep its old value unless you use this flag
    to change the setting.
    """
  group.add_argument('--memory', type=str, help=memory_help_text, required=True)
  group.add_argument('--cpu', help=cpu_help_text)


def ParseMemoryStrToNumBytes(binary_size):
  """Parse binary size to number of bytes.

  Args:
    binary_size: str, memory with size suffix

  Returns:
    num_bytes: int, the number of bytes
  """

  binary_size_parser = arg_parsers.BinarySize(
      suggested_binary_size_scales=['KB', 'MB', 'MiB', 'GB', 'GiB'],
      default_unit='MB',
  )
  return binary_size_parser(binary_size)


def ValidateV1TimeoutFlag(args):
  if args.timeout and args.timeout > 540:
    raise exceptions.InvalidArgumentException(
        '--timeout',
        'value must be less than or equal to 540s for 1st gen functions;'
        ' received: {}s'.format(args.timeout),
    )


def AddFunctionTimeoutFlag(parser):
  """Add flag for specifying function timeout to the parser.

  Args:
    parser: the argparse parser for the command.

  Returns:
    None
  """

  help_text = """\
      The function execution timeout, e.g. 30s for 30 seconds. Defaults to
      original value for existing function or 60 seconds for new functions.

      For GCF 1st gen functions, cannot be more than 540s.

      For GCF 2nd gen functions, cannot be more than 3600s.

      See $ gcloud topic datetimes for information on duration formats."""

  parser.add_argument(
      '--timeout', help=help_text, type=arg_parsers.Duration(lower_bound='1s')
  )


def AddFunctionRetryFlag(parser):
  """Add flag for specifying function retry behavior to the parser."""
  parser.add_argument(
      '--retry',
      help=(
          'If specified, then the function will be retried in case of a '
          'failure.'
      ),
      action='store_true',
  )


def AddAllowUnauthenticatedFlag(parser):
  """Add the --allow-unauthenticated flag."""
  parser.add_argument(
      '--allow-unauthenticated',
      action=arg_parsers.StoreTrueFalseAction,
      help=(
          'If set, makes this a public function. This will allow all '
          'callers, without checking authentication.'
      ),
  )


def AddServeAllTrafficLatestRevisionFlag(parser):
  help_text = (
      'If specified, latest function revision will be served all traffic.'
  )
  parser.add_argument(
      '--serve-all-traffic-latest-revision',
      action='store_true',
      default=False,
      help=help_text,
  )


def AddBuildpackStackFlag(parser):
  """Add flag for specifying function memory to the parser."""
  # TODO(b/211892283): refer to Buildpack stacks guide, once it is available
  help_text = """\
  Specifies one of the Google provided buildpack stacks.
  """
  parser.add_argument('--buildpack-stack', type=str, help=help_text)


def AddGen2Flag(
    parser, operates_on_existing_function=True, hidden=False, allow_v2=False
):
  """Add the --gen2 flag."""
  help_text = (
      'If enabled, this command will use Cloud Functions (Second generation).'
      ' If disabled with `--no-gen2`, Cloud Functions (First generation) will'
      ' be used. If not specified, the value of this flag will be taken from'
      ' the `functions/gen2` configuration property.'
  )
  if operates_on_existing_function:
    help_text += (
        ' If the `functions/gen2` configuration property is not set, defaults'
        ' to looking up the given function and using its generation.'
    )
  if allow_v2:
    help_text += (
        ' This command could conflict with `--v2`. If specified `--gen2`'
        ' with `--no-v2`, or `--no-gen2` with `--v2`, Second generation will be'
        ' used.'
    )
  parser.add_argument(
      '--gen2',
      default=False,
      action=actions.StoreBooleanProperty(properties.VALUES.functions.gen2),
      help=help_text,
      hidden=hidden,
  )


def ShouldUseGen2():
  """Returns whether 2nd gen should be used for Cloud Functions."""
  return bool(properties.VALUES.functions.gen2.GetBool())


def ShouldUseGen1():
  """Returns whether 1st gen should be used for Cloud Functions."""
  return (
      properties.VALUES.functions.gen2.IsExplicitlySet() and not ShouldUseGen2()
  )


def ShouldEnsureAllUsersInvoke(args):
  return args.allow_unauthenticated


def ShouldDenyAllUsersInvoke(args):
  return (
      args.IsSpecified('allow_unauthenticated')
      and not args.allow_unauthenticated
  )


def AddV2Flag(parser):
  """Add the --v2 flag."""
  help_text = (
      'If specified, this command will use Cloud Functions v2 APIs and return'
      ' the result in the v2 format (See'
      ' https://cloud.google.com/functions/docs/reference/rest/v2/projects.locations.functions#Function).'
      ' If not specified, 1st gen and 2nd gen functions will use v1 and v2 APIs'
      ' respectively and return the result in the corresponding format (For v1'
      ' format, see'
      ' https://cloud.google.com/functions/docs/reference/rest/v1/projects.locations.functions#resource:-cloudfunction).'
      ' This command conflicts with `--no-gen2`. If specified with this'
      ' combination, v2 APIs will be used.'
  )
  parser.add_argument(
      '--v2',
      action='store_true',
      default=None,
      help=help_text,
  )


def AddSourceFlag(parser):
  """Add flag for specifying function source code to the parser."""
  parser.add_argument(
      '--source',
      completer=DirectoriesCompleter,
      help="""\
      Location of source code to deploy.

      Location of the source can be one of the following three options:

      * Source code in Google Cloud Storage (must be a `.zip` archive),
      * Reference to source repository or,
      * Local filesystem path (root directory of function source).

      Note that, depending on your runtime type, Cloud Functions will look
      for files with specific names for deployable functions. For Node.js,
      these filenames are `index.js` or `function.js`. For Python, this is
      `main.py`.

      If you do not specify the `--source` flag:

      * The current directory will be used for new function deployments.
      * If the function was previously deployed using a local filesystem path,
      then the function's source code will be updated using the current
      directory.
      * If the function was previously deployed using a Google Cloud Storage
      location or a source repository, then the function's source code will not
      be updated.

      The value of the flag will be interpreted as a Cloud Storage location, if
      it starts with `gs://`.

      The value will be interpreted as a reference to a source repository, if it
      starts with `https://`.

      Otherwise, it will be interpreted as the local filesystem path. When
      deploying source from the local filesystem, this command skips files
      specified in the `.gcloudignore` file (see `gcloud topic gcloudignore` for
      more information). If the `.gcloudignore` file doesn't exist, the command
      will try to create it.

      The minimal source repository URL is:
      `https://source.developers.google.com/projects/${PROJECT}/repos/${REPO}`

      By using the URL above, sources from the root directory of the
      repository on the revision tagged `master` will be used.

      If you want to deploy from a revision different from `master`, append one
      of the following three sources to the URL:

      * `/revisions/${REVISION}`,
      * `/moveable-aliases/${MOVEABLE_ALIAS}`,
      * `/fixed-aliases/${FIXED_ALIAS}`.

      If you'd like to deploy sources from a directory different from the root,
      you must specify a revision, a moveable alias, or a fixed alias, as above,
      and append `/paths/${PATH_TO_SOURCES_DIRECTORY}` to the URL.

      Overall, the URL should match the following regular expression:

      ```
      ^https://source\\.developers\\.google\\.com/projects/
      (?<accountId>[^/]+)/repos/(?<repoName>[^/]+)
      (((/revisions/(?<commit>[^/]+))|(/moveable-aliases/(?<branch>[^/]+))|
      (/fixed-aliases/(?<tag>[^/]+)))(/paths/(?<path>.*))?)?$
      ```

      An example of a validly formatted source repository URL is:

      ```
      https://source.developers.google.com/projects/123456789/repos/testrepo/
      moveable-aliases/alternate-branch/paths/path-to=source
      ```

      """,
  )


def AddStageBucketFlag(parser):
  """Add flag for specifying stage bucket to the parser."""
  parser.add_argument(
      '--stage-bucket',
      help=(
          "When deploying a function from a local directory, this flag's "
          'value is the name of the Google Cloud Storage bucket in which '
          'source code will be stored. Note that if you set the '
          '`--stage-bucket` flag when deploying a function, you will need to '
          'specify `--source` or `--stage-bucket` in subsequent deployments '
          'to update your source code. To use this flag successfully, the '
          'account in use must have permissions to write to this bucket. For '
          'help granting access, refer to this guide: '
          'https://cloud.google.com/storage/docs/access-control/'
      ),
      type=api_util.ValidateAndStandarizeBucketUriOrRaise,
  )


def AddRuntimeFlag(parser, hidden=False):
  parser.add_argument(
      '--runtime',
      help="""\
          Runtime in which to run the function.

          Required when deploying a new function; optional when updating
          an existing function.

          For a list of available runtimes, run `gcloud functions runtimes list`.
          """,
      hidden=hidden,
  )


def GetVpcConnectorResourceSpec():
  return concepts.ResourceSpec(
      'vpcaccess.projects.locations.connectors',
      resource_name='connector',
      disable_auto_completers=False,
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
      locationsId=RegionAttributeConfig(),
      connectorsId=VpcConnectorAttributeConfig(),
  )


def AddVPCConnectorMutexGroup(parser):
  """Add flag for specifying VPC connector to the parser."""
  mutex_group = parser.add_group(mutex=True)
  resource = presentation_specs.ResourcePresentationSpec(
      '--vpc-connector',
      GetVpcConnectorResourceSpec(),
      """\
        The VPC Access connector that the function can connect to. It can be
        either the fully-qualified URI, or the short name of the VPC Access
        connector resource. If the short name is used, the connector must
        belong to the same project. The format of this field is either
        `projects/${PROJECT}/locations/${LOCATION}/connectors/${CONNECTOR}`
        or `${CONNECTOR}`, where `${CONNECTOR}` is the short name of the VPC
        Access connector.
      """,
      group=mutex_group,
      # This hides the region flag for the connector resource.
      flag_name_overrides={'region': ''},
  )

  concept_parsers.ConceptParser(
      [resource],
      # This configures the fallthrough from the vpc-connector's region to
      # the primary flag for the function's region.
      command_level_fallthroughs={'--vpc-connector.region': ['--region']},
  ).AddToParser(parser)

  mutex_group.add_argument(
      '--clear-vpc-connector',
      action='store_true',
      help="""\
        Clears the VPC connector field.
      """,
  )


def _AddVpcNetworkFlag(parser):
  """Add flags for setting VPC network."""
  parser.add_argument(
      '--network',
      help=(
          'The VPC network that the Cloud Function will be able to send'
          ' traffic to. If --subnet is also specified, subnet must be a'
          ' subnetwork of the network specified by this --network flag. To'
          ' clear existing VPC network settings, use --clear-network.'
      ),
  )


def _AddVpcSubnetFlag(parser):
  """Add flags for setting VPC subnetwork."""
  parser.add_argument(
      '--subnet',
      help=(
          'The VPC subnetwork that the Cloud Function will get IPs from. The'
          ' subnetwork must be `/26` or larger. If --network is also specified,'
          ' subnet must be a subnetwork of the network specified by the'
          ' --network flag. If --network is not specified, network will be'
          ' looked up from this subnetwork. To clear existing VPC network'
          ' settings, use --clear-network.'
      ),
  )


def _AddVpcNetworkTagsFlag(parser):
  """Add flags for setting VPC network tags."""
  parser.add_argument(
      '--network-tags',
      type=arg_parsers.ArgList(),
      metavar='TAGS',
      help=(
          'Applies the given network tags (comma separated) to the '
          'Cloud Function. '
          'To clear existing tags, use --clear-network-tags.'
      ),
  )


def _AddClearVpcNetworkTagsFlag(parser):
  """Add flags for clearing VPC network tags."""
  parser.add_argument(
      '--clear-network-tags',
      action='store_true',
      help=(
          'Clears all existing network tags from the Cloud Function.'
      ),
  )


def _AddDirectVpcEgressFlag(parser):
  """Add flags for setting Direct VPC egress settings."""
  direct_vpc_egress_arg = base.ChoiceArgument(
      '--direct-vpc-egress',
      choices=[x.lower() for x in EGRESS_SETTINGS],
      help_str=(
          'Specify which of the outbound traffic to send through Direct VPC'
          ' egress. Configuring Direct VPC network is required to use this'
          ' flag.'
      ),
  )
  direct_vpc_egress_arg.AddToParser(parser)


def _AddNetworkInterfaceGroupFlags(parser):
  """Add flag for specifying Direct VPC network interface to the parser."""
  group = parser.add_argument_group(
      'Direct VPC egress setting flags group.'
  )
  _AddVpcNetworkFlag(group)
  _AddVpcSubnetFlag(group)

  tags_group = group.add_group(mutex=True)
  _AddVpcNetworkTagsFlag(tags_group)
  _AddClearVpcNetworkTagsFlag(tags_group)


def _AddClearVpcNetworkFlag(parser):
  """Add flags for clearing VPC network."""
  parser.add_argument(
      '--clear-network',
      action='store_true',
      help=(
          'Disconnect this Cloud Function from the Direct VPC network it is'
          ' connected to.'
      ),
  )


def AddAllDirectVpcFlags(parser):
  """Add flags for all Direct VPC network related settings."""
  ni_group = parser.add_group(mutex=True)
  _AddNetworkInterfaceGroupFlags(ni_group)
  _AddClearVpcNetworkFlag(ni_group)

  _AddDirectVpcEgressFlag(parser)


def AddBuildWorkerPoolMutexGroup(parser):
  """Add flag for specifying Build Worker Pool to the parser."""
  mutex_group = parser.add_group(mutex=True)
  mutex_group.add_argument(
      '--build-worker-pool',
      help="""\
        Name of the Cloud Build Custom Worker Pool that should be used to build
        the function. The format of this field is
        `projects/${PROJECT}/locations/${LOCATION}/workerPools/${WORKERPOOL}`
        where ${PROJECT} is the project id and ${LOCATION} is the location where
        the worker pool is defined and ${WORKERPOOL} is the short name of the
        worker pool.
      """,
  )
  mutex_group.add_argument(
      '--clear-build-worker-pool',
      action='store_true',
      help="""\
        Clears the Cloud Build Custom Worker Pool field.
      """,
  )


def AddEntryPointFlag(parser):
  """Add flag for specifying entry point to the parser."""
  parser.add_argument(
      '--entry-point',
      help="""\
      Name of a Google Cloud Function (as defined in source code) that will
      be executed. Defaults to the resource name suffix (ID of the function), if
       not specified.
""",
  )


def AddMaxInstancesFlag(parser):
  """Add flag for specifying the max instances for a function."""
  mutex_group = parser.add_group(mutex=True)
  mutex_group.add_argument(
      '--max-instances',
      type=arg_parsers.BoundedInt(lower_bound=1),
      help="""\
        Sets the maximum number of instances for the function. A function
        execution that would exceed max-instances times out.
      """,
  )
  mutex_group.add_argument(
      '--clear-max-instances',
      action='store_true',
      help="""\
        Clears the maximum instances setting for the function.

        If it's any 2nd gen function or a 1st gen HTTP function, this flag sets
        maximum instances  to 0, which means there is no limit to maximum
        instances. If it's an event-driven 1st gen function, this flag sets
        maximum instances to 3000, which is the default value for 1st gen
        functions.
      """,
  )


def AddMinInstancesFlag(parser):
  """Add flag for specifying the min instances for a function."""
  mutex_group = parser.add_group(mutex=True)
  mutex_group.add_argument(
      '--min-instances',
      type=arg_parsers.BoundedInt(lower_bound=0),
      help="""\
        Sets the minimum number of instances for the function. This is helpful
        for reducing cold start times. Defaults to zero.
      """,
  )
  mutex_group.add_argument(
      '--clear-min-instances',
      action='store_true',
      help="""\
        Clears the minimum instances setting for the function.
      """,
  )


def AddTriggerFlagGroup(parser):
  """Add arguments specifying functions trigger to the parser.

  Args:
    parser: the argparse parser for the command.
  """
  trigger_group = parser.add_mutually_exclusive_group(help="""\
      If you don't specify a trigger when deploying an update to an existing
      function it will keep its current trigger. You must specify one of the
      following when deploying a new function:
      - `--trigger-topic`,
      - `--trigger-bucket`,
      - `--trigger-http`,
      - `--trigger-event` AND `--trigger-resource`,
      - `--trigger-event-filters` and optionally `--trigger-event-filters-path-pattern`.
      """)
  trigger_group.add_argument(
      '--trigger-topic',
      help=(
          'Name of Pub/Sub topic. Every message published in this topic '
          'will trigger function execution with message contents passed as '
          'input data. Note that this flag does not accept the format of '
          'projects/PROJECT_ID/topics/TOPIC_ID. Use this flag to specify the '
          'final element TOPIC_ID. The PROJECT_ID will be read from the '
          'active configuration.'
      ),
      type=api_util.ValidatePubsubTopicNameOrRaise,
  )
  trigger_group.add_argument(
      '--trigger-bucket',
      help=(
          'Google Cloud Storage bucket name. Trigger the function when an '
          'object is created or overwritten in the specified Cloud Storage '
          'bucket.'
      ),
      type=api_util.ValidateAndStandarizeBucketUriOrRaise,
  )
  trigger_group.add_argument(
      '--trigger-http',
      action='store_true',
      help="""\
      Function will be assigned an endpoint, which you can view by using
      the `describe` command. Any HTTP request (of a supported type) to the
      endpoint will trigger function execution. Supported HTTP request
      types are: POST, PUT, GET, DELETE, and OPTIONS.""",
  )
  eventarc_trigger_group = trigger_group.add_argument_group()
  concept_parsers.ConceptParser(
      [
          presentation_specs.ResourcePresentationSpec(
              '--trigger-channel',
              eventarc_flags.ChannelResourceSpec(),
              """\
              The channel to use in the trigger for third-party event sources.
              """,
              flag_name_overrides={'location': ''},
              group=eventarc_trigger_group,
              hidden=True,
          )
      ],
      command_level_fallthroughs={
          '--trigger-channel.location': ['--trigger-location'],
      },
  ).AddToParser(parser)
  eventarc_trigger_group.add_argument(
      '--trigger-event-filters',
      type=arg_parsers.ArgDict(),
      action=arg_parsers.UpdateAction,
      metavar='ATTRIBUTE=VALUE',
      help="""\
      The Eventarc matching criteria for the trigger. The criteria can be
      specified either as a single comma-separated argument or as multiple
      arguments. The filters must include the ``type'' attribute, as well as any
      other attributes that are expected for the chosen type.
      """,
  )
  eventarc_trigger_group.add_argument(
      '--trigger-event-filters-path-pattern',
      type=arg_parsers.ArgDict(),
      action=arg_parsers.UpdateAction,
      metavar='ATTRIBUTE=PATH_PATTERN',
      help="""\
      The Eventarc matching criteria for the trigger in path pattern format.
      The criteria can be specified as a single comma-separated argument or as
      multiple arguments.

      The provided attribute/value pair will be used with the
      `match-path-pattern` operator to configure the trigger, see
      https://cloud.google.com/eventarc/docs/reference/rest/v1/projects.locations.triggers#eventfilter
      and https://cloud.google.com/eventarc/docs/path-patterns for more details about on
      how to construct path patterns.

      For example, to filter on events for Compute Engine VMs in a given zone:
      `--trigger-event-filters-path-pattern=resourceName='/projects/*/zones/us-central1-a/instances/*'""",
  )

  trigger_provider_spec_group = trigger_group.add_argument_group()
  # check later as type of applicable input depends on options above
  trigger_provider_spec_group.add_argument(
      '--trigger-event',
      metavar='EVENT_TYPE',
      help=(
          'Specifies which action should trigger the function. For a '
          'list of acceptable values, call '
          '`gcloud functions event-types list`.'
      ),
  )
  trigger_provider_spec_group.add_argument(
      '--trigger-resource',
      metavar='RESOURCE',
      help=(
          'Specifies which resource from `--trigger-event` is being '
          'observed. E.g. if `--trigger-event` is  '
          '`providers/cloud.storage/eventTypes/object.change`, '
          '`--trigger-resource` must be a bucket name. For a list of '
          'expected resources, call '
          '`gcloud functions event-types list`.'
      ),
  )


class LocationsCompleter(completers.ListCommandCompleter):

  def __init__(self, **kwargs):
    super(LocationsCompleter, self).__init__(
        collection=LOCATIONS_COLLECTION,
        list_command='alpha functions regions list --uri',
        **kwargs,
    )


class RegionFallthrough(deps.PropertyFallthrough):
  """Custom fallthrough for region dependent on GCF generation.

  For GCF gen1 this falls back to the functions/region property.

  For GCF gen2 the property fallback is only used if it is explicitly set.
  Otherwise the region is prompted for.
  """

  def __init__(self, release_track=base.ReleaseTrack.ALPHA):
    super(RegionFallthrough, self).__init__(properties.VALUES.functions.region)
    self.release_track = release_track

  def _Call(self, parsed_args):
    use_gen1 = not ShouldUseGen2()
    if use_gen1 or self.property.IsExplicitlySet():
      return super(RegionFallthrough, self)._Call(parsed_args)

    if not console_io.CanPrompt():
      raise exceptions.RequiredArgumentException(
          'region',
          (
              'You must specify a region. Either use the flag `--region` or set'
              ' the functions/region property.'
          ),
      )

    client = client_v2.FunctionsClient(self.release_track)
    regions = [l.locationId for l in client.ListRegions()]
    idx = console_io.PromptChoice(regions, message='Please specify a region:\n')
    region = regions[idx]
    log.status.Print(
        'To make this the default region, run '
        '`gcloud config set functions/region {}`.\n'.format(region)
    )

    return region


def AddRegionFlag(parser, help_text):
  parser.add_argument(
      '--region',
      help=help_text,
      completer=LocationsCompleter,
      action=actions.StoreProperty(properties.VALUES.functions.region),
  )


def RegionAttributeConfig():
  return concepts.ResourceParameterAttributeConfig(
      name='region',
      help_text=(
          'The Cloud region for the {resource}. Overrides the default '
          '`functions/region` property value for this command invocation.'
      ),
      completer=LocationsCompleter,
      fallthroughs=[RegionFallthrough()],
  )


def AddTriggerLocationFlag(parser):
  """Add flag for specifying trigger location to the parser."""
  parser.add_argument(
      '--trigger-location',
      help=(
          'The location of the trigger, which must be a region or multi-'
          'region where the relevant events originate.'
      ),
      completer=LocationsCompleter,
  )


def FunctionAttributeConfig():
  return concepts.ResourceParameterAttributeConfig(
      name='function',
      help_text='The name of the {resource}.',
      value_type=api_util.ValidateFunctionNameOrRaise,
  )


def VpcConnectorAttributeConfig():
  return concepts.ResourceParameterAttributeConfig(
      name='connector',
      help_text='The name of the {resource}.',
  )


def GetFunctionResourceSpec():
  return concepts.ResourceSpec(
      'cloudfunctions.projects.locations.functions',
      resource_name='function',
      disable_auto_completers=False,
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
      locationsId=RegionAttributeConfig(),
      functionsId=FunctionAttributeConfig(),
  )


def AddFunctionResourceArg(
    parser, verb='', help_text_override='', required=True
):
  """Adds a Cloud Function resource argument.

  NOTE: May be used only if it's the only resource arg in the command.
  Args:
    parser: the argparse parser for the command.
    verb: (Optional) str, the verb to describe the resource, such as 'to
      update'.
    help_text_override: (Optional)str, the help text to use for the resource
      argument. If override is providded, verb will be ignored.
    required: bool, whether the argument is required.
  """
  help_text = help_text_override or 'The Cloud Function name {}.'.format(verb)
  concept_parsers.ConceptParser.ForResource(
      'NAME',
      GetFunctionResourceSpec(),
      help_text,
      required=required,
  ).AddToParser(parser)


def AddServiceAccountFlag(parser):
  parser.add_argument(
      '--service-account',
      help="""\
      The email address of the IAM service account associated with the
      function at runtime. The service account represents the identity of the
      running function, and determines what permissions the function has.

      If not provided, the function will use the project's default service
      account for Compute Engine.
      """,
  )


def AddRunServiceAccountFlag(parser):
  parser.add_argument(
      '--run-service-account',
      help="""\
      The email address of the IAM service account associated with the Cloud
      Run service for the function. The service account represents the identity
      of the running function, and determines what permissions the function
      has.

      If not provided, the function will use the project's default service
      account for Compute Engine.
      """,
  )


def AddTriggerServiceAccountFlag(parser):
  parser.add_argument(
      '--trigger-service-account',
      help="""\
      The email address of the IAM service account associated with the Eventarc
      trigger for the function. This is used for authenticated invocation.

      If not provided, the function will use the project's default service
      account for Compute Engine.
      """,
  )


def AddDataFlag(parser):
  parser.add_argument(
      '--data',
      help="""JSON string with data that will be passed to the function.""",
      type=_ValidateJsonOrRaiseDataError,
  )


def AddCloudEventsFlag(parser):
  parser.add_argument(
      '--cloud-event',
      help="""
      JSON encoded string with a CloudEvent in structured content mode.

      Mutually exclusive with --data flag.

      Use for Cloud Functions 2nd Gen CloudEvent functions. The CloudEvent
      object will be sent to your function as a binary content mode message with
      the top-level 'data' field set as the HTTP body and all other JSON fields
      sent as HTTP headers.
      """,
      type=_ValidateJsonOrRaiseCloudEventError,
  )


def AddGcloudHttpTimeoutFlag(parser):
  """Add flag for specifying gcloud timeout to the parser."""

  help_text = """\
      The gcloud command timeout, e.g. 30s for 30 seconds. Defaults to the
      function execution timeout duration of the function.

      See $ gcloud topic datetimes for information on duration formats."""
  parser.add_argument(
      '--timeout',
      help=help_text,
      type=arg_parsers.Duration(lower_bound='1s'),
      hidden=True,
  )


def AddIAMPolicyFileArg(parser):
  parser.add_argument(
      'policy_file',
      metavar='POLICY_FILE',
      help=(
          'Path to a local JSON or YAML formatted file '
          'containing a valid policy.'
      ),
  )


def AddIgnoreFileFlag(parser):
  parser.add_argument(
      '--ignore-file',
      help=(
          'Override the .gcloudignore file in the source directory and use the'
          ' specified file instead. By default, the source directory is your'
          ' current directory. Note that it could be changed by the --source'
          ' flag, in which case your .gcloudignore file will be searched in the'
          ' overridden directory.  For example, `--ignore-file=.mygcloudignore`'
          ' combined with `--source=./mydir` would point to'
          ' `./mydir/.mygcloudignore`'
      ),
  )


# Flags for Runtime Updates
def AddRuntimeUpdatePolicy(parser, track):
  """Adds flags for selecting the runtime update policy."""
  if track in (
      base.ReleaseTrack.ALPHA,
      base.ReleaseTrack.BETA,
      base.ReleaseTrack.GA,
  ):
    parser.add_argument(
        '--runtime-update-policy',
        choices=RUNTIME_UPDATE_POLICY_MAPPING.values(),
        help="""\
        Runtime update policy for the function being deployed. The option
        `automatic` is used by default.
      """,
    )


# Flags for Artifact Registry
def AddDockerRegistryFlags(parser):
  """Adds flags for selecting the Docker registry type for Cloud Function."""
  docker_registry_arg = base.ChoiceArgument(
      '--docker-registry',
      choices=sorted(DOCKER_REGISTRY_MAPPING.values()),
      help_str="""\
        Docker Registry to use for storing the function's Docker images.
        The option `artifact-registry` is used by default.
      """,
      action=actions.DeprecationAction('--docker-registry', warn="""\
        With the general transition from Container Registry to
        Artifact Registry, the option to specify docker registry is deprecated.
        All container image storage and management will automatically
        transition to Artifact Registry.
        For more information, see
        https://cloud.google.com/artifact-registry/docs/transition/transition-from-gcr""")
  )
  docker_registry_arg.AddToParser(parser)


# Flags for CMEK
def AddKMSKeyFlags(parser):
  """Adds flags for configuring the CMEK key."""
  kmskey_group = parser.add_group(mutex=True)
  kmskey_group.add_argument(
      '--kms-key',
      type=arg_parsers.RegexpValidator(
          _KMS_KEY_NAME_PATTERN, _KMS_KEY_NAME_ERROR
      ),
      help="""\
        Sets the user managed KMS crypto key used to encrypt the Cloud Function
        and its resources.

        The KMS crypto key name should match the pattern
        `projects/${PROJECT}/locations/${LOCATION}/keyRings/${KEYRING}/cryptoKeys/${CRYPTOKEY}`
        where ${PROJECT} is the project, ${LOCATION} is the location of the key
        ring, and ${KEYRING} is the key ring that contains the ${CRYPTOKEY}
        crypto key.

        If this flag is set, then a Docker repository created in Artifact
        Registry must be specified using the `--docker-repository` flag and the
        repository must be encrypted using the `same` KMS key.
      """,
  )
  kmskey_group.add_argument(
      '--clear-kms-key',
      action='store_true',
      help="""\
        Clears the KMS crypto key used to encrypt the function.
      """,
  )


def AddDockerRepositoryFlags(parser):
  """Adds flags for configuring the Docker repository for Cloud Function."""
  kmskey_group = parser.add_group(mutex=True)
  kmskey_group.add_argument(
      '--docker-repository',
      type=arg_parsers.RegexpValidator(
          _DOCKER_REPOSITORY_NAME_PATTERN, _DOCKER_REPOSITORY_NAME_ERROR
      ),
      help="""\
        Sets the Docker repository to be used for storing the Cloud Function's
        Docker images while the function is being deployed. `DOCKER_REPOSITORY`
        must be an Artifact Registry Docker repository present in the `same`
        project and location as the Cloud Function.

        **Preview:** for 2nd gen functions, a Docker Artifact registry
        repository in a different project and/or location may be used.
        Additional requirements apply, see
        https://cloud.google.com/functions/docs/building#image_registry

        The repository name should match one of these patterns:

        * `projects/${PROJECT}/locations/${LOCATION}/repositories/${REPOSITORY}`,
        * `{LOCATION}-docker.pkg.dev/{PROJECT}/{REPOSITORY}`.

        where `${PROJECT}` is the project, `${LOCATION}` is the location of the
        repository and `${REPOSITORY}` is a valid repository ID.
      """,
  )
  kmskey_group.add_argument(
      '--clear-docker-repository',
      action='store_true',
      help="""\
        Clears the Docker repository configuration of the function.
      """,
  )


def AddConcurrencyFlag(parser):
  parser.add_argument(
      '--concurrency',
      type=arg_parsers.BoundedInt(lower_bound=1, upper_bound=1000),
      help=(
          'Set the maximum number of concurrent requests allowed per'
          ' container instance. Leave concurrency unspecified to receive the'
          ' server default value.'
      ),
  )


def AddUpgradeFlags(parser, release_track=base.ReleaseTrack.GA):
  """Adds upgrade related function flags."""
  description = 'Upgrade a 1st gen Cloud Function to a Cloud Run Function.'
  commit_flag_help = '- `--commit` and optionally `--skip-detach`'

  upgrade_group = parser.add_group(
      mutex=True,
      help="""\
      {}
      You must specify one of the following flags:
      - `--setup-config` and optionally `--trigger-service-account`,
      - `--redirect-traffic`,
      - `--rollback-traffic`,
      {},
      - `--abort`.
  """.format(description, commit_flag_help),
  )

  setup_config_group = upgrade_group.add_argument_group()
  setup_config_group.add_argument(
      '--setup-config',
      action='store_true',
      help=(
          'Sets up the function upgrade config by creating a 2nd gen copy of'
          " the function's code and configuration."
      ),
  )
  AddTriggerServiceAccountFlag(setup_config_group)
  if release_track in (base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA):
    AddRuntimeFlag(setup_config_group)
    setup_config_group.add_argument(
        '--max-instances',
        type=arg_parsers.BoundedInt(lower_bound=1),
        help="""\
          Sets the maximum number of instances for the function. A function
          execution that would exceed max-instances times out.
        """,
    )

  upgrade_group.add_argument(
      '--redirect-traffic',
      action='store_true',
      help='Redirects production traffic to the 2nd gen copy of the function.',
  )
  upgrade_group.add_argument(
      '--rollback-traffic',
      action='store_true',
      help=(
          'Rolls back production traffic to the original 1st gen copy of the'
          ' function. The 2nd gen copy will still be available for testing.'
      ),
  )

  commit_group = upgrade_group.add_argument_group()
  commit_group.add_argument(
      '--commit',
      action='store_true',
      help=(
          'Finishes the upgrade process and permanently deletes the original'
          ' 1st gen copy of the function.'
      ),
  )
  if release_track in (base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA):
    commit_group.add_argument(
        '--skip-detach',
        action='store_true',
        hidden=False,
        help=(
            'The function will not be detached when committing the upgrade to'
            ' allow continued use of the Cloud Functions v2 API. You can detach'
            ' the function to Cloud Run afterward'
            ' https://docs.cloud.google.com/run/docs/functions/comparison#detach_your_function.'
            ' This flag is only valid when used with the --commit flag.'
        ),
    )

  upgrade_group.add_argument(
      '--abort',
      action='store_true',
      help=(
          'Undoes all steps of the upgrade process done so far. All traffic'
          ' will point to the original 1st gen function copy and the 2nd gen'
          ' function copy will be deleted.'
      ),
  )


def _ValidateJsonOrRaiseDataError(data):
  return _ValidateJsonOrRaiseError(data, '--data')


def _ValidateJsonOrRaiseCloudEventError(data):
  return _ValidateJsonOrRaiseError(data, '--cloud-event')


def _ValidateJsonOrRaiseError(data, arg_name):
  """Checks validity of json string or raises an InvalidArgumentException."""
  try:
    json.loads(data)
    return data
  except ValueError as e:
    raise exceptions.InvalidArgumentException(
        arg_name, 'Is not a valid JSON: ' + six.text_type(e)
    )


def AddBuildServiceAccountFlag(parser):
  """Adds flags for configuring the build service account for Cloud Function."""
  mutex_group = parser.add_group(mutex=True)
  mutex_group.add_argument(
      '--build-service-account',
      help="""\
          IAM service account whose credentials will be used for the build step.
          Must be of the format projects/${PROJECT_ID}/serviceAccounts/${ACCOUNT_EMAIL_ADDRESS}.

          If not provided, the function will use the project's default
          service account for Cloud Build.
      """,
  )
  mutex_group.add_argument(
      '--clear-build-service-account',
      action='store_true',
      help="""\
          Clears the build service account field.
      """,
  )


def AddBinaryAuthorizationMutexGroup(parser):
  """Add flag for specifying Binary Authorization Policy to the parser."""
  mutex_group = parser.add_group(mutex=True)
  mutex_group.add_argument(
      '--binary-authorization',
      help="""\
        Name of the Binary Authorization policy that the function image should
        be checked against when deploying to Cloud Run.

        Example: default

        The flag is only applicable to 2nd gen functions.
      """,
  )
  mutex_group.add_argument(
      '--clear-binary-authorization',
      action='store_true',
      help="""\
        Clears the Binary Authorization policy field.
      """,
  )



# Copyright 2016 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for transitioning from oauth2client to google-auth.

.. warning::
    This module is private as it is intended to assist first-party downstream
    clients with the transition from oauth2client to google-auth.
"""

from __future__ import absolute_import

from google.auth import _helpers
import google.auth.app_engine
import google.auth.compute_engine
import google.oauth2.credentials
import google.oauth2.service_account

try:
    import oauth2client.client  # type: ignore
    import oauth2client.contrib.gce  # type: ignore
    import oauth2client.service_account  # type: ignore
except ImportError as caught_exc:
    raise ImportError("oauth2client is not installed.") from caught_exc

try:
    import oauth2client.contrib.appengine  # type: ignore

    _HAS_APPENGINE = True
except ImportError:
    _HAS_APPENGINE = False


_CONVERT_ERROR_TMPL = "Unable to convert {} to a google-auth credentials class."


def _convert_oauth2_credentials(credentials):
    """Converts to :class:`google.oauth2.credentials.Credentials`.

    Args:
        credentials (Union[oauth2client.client.OAuth2Credentials,
            oauth2client.client.GoogleCredentials]): The credentials to
            convert.

    Returns:
        google.oauth2.credentials.Credentials: The converted credentials.
    """
    new_credentials = google.oauth2.credentials.Credentials(
        token=credentials.access_token,
        refresh_token=credentials.refresh_token,
        token_uri=credentials.token_uri,
        client_id=credentials.client_id,
        client_secret=credentials.client_secret,
        scopes=credentials.scopes,
    )

    new_credentials._expires = credentials.token_expiry

    return new_credentials


def _convert_service_account_credentials(credentials):
    """Converts to :class:`google.oauth2.service_account.Credentials`.

    Args:
        credentials (Union[
            oauth2client.service_account.ServiceAccountCredentials,
            oauth2client.service_account._JWTAccessCredentials]): The
            credentials to convert.

    Returns:
        google.oauth2.service_account.Credentials: The converted credentials.
    """
    info = credentials.serialization_data.copy()
    info["token_uri"] = credentials.token_uri
    return google.oauth2.service_account.Credentials.from_service_account_info(info)


def _convert_gce_app_assertion_credentials(credentials):
    """Converts to :class:`google.auth.compute_engine.Credentials`.

    Args:
        credentials (oauth2client.contrib.gce.AppAssertionCredentials): The
            credentials to convert.

    Returns:
        google.oauth2.service_account.Credentials: The converted credentials.
    """
    return google.auth.compute_engine.Credentials(
        service_account_email=credentials.service_account_email
    )


def _convert_appengine_app_assertion_credentials(credentials):
    """Converts to :class:`google.auth.app_engine.Credentials`.

    Args:
        credentials (oauth2client.contrib.app_engine.AppAssertionCredentials):
            The credentials to convert.

    Returns:
        google.oauth2.service_account.Credentials: The converted credentials.
    """
    # pylint: disable=invalid-name
    return google.auth.app_engine.Credentials(
        scopes=_helpers.string_to_scopes(credentials.scope),
        service_account_id=credentials.service_account_id,
    )


_CLASS_CONVERSION_MAP = {
    oauth2client.client.OAuth2Credentials: _convert_oauth2_credentials,
    oauth2client.client.GoogleCredentials: _convert_oauth2_credentials,
    oauth2client.service_account.ServiceAccountCredentials: _convert_service_account_credentials,
    oauth2client.service_account._JWTAccessCredentials: _convert_service_account_credentials,
    oauth2client.contrib.gce.AppAssertionCredentials: _convert_gce_app_assertion_credentials,
}

if _HAS_APPENGINE:  # pragma: no cover
    _CLASS_CONVERSION_MAP[
        oauth2client.contrib.appengine.AppAssertionCredentials
    ] = _convert_appengine_app_assertion_credentials


def convert(credentials):
    """Convert oauth2client credentials to google-auth credentials.

    This class converts:

    - :class:`oauth2client.client.OAuth2Credentials` to
      :class:`google.oauth2.credentials.Credentials`.
    - :class:`oauth2client.client.GoogleCredentials` to
      :class:`google.oauth2.credentials.Credentials`.
    - :class:`oauth2client.service_account.ServiceAccountCredentials` to
      :class:`google.oauth2.service_account.Credentials`.
    - :class:`oauth2client.service_account._JWTAccessCredentials` to
      :class:`google.oauth2.service_account.Credentials`.
    - :class:`oauth2client.contrib.gce.AppAssertionCredentials` to
      :class:`google.auth.compute_engine.Credentials`.
    - :class:`oauth2client.contrib.appengine.AppAssertionCredentials` to
      :class:`google.auth.app_engine.Credentials`.

    Returns:
        google.auth.credentials.Credentials: The converted credentials.

    Raises:
        ValueError: If the credentials could not be converted.
    """

    credentials_class = type(credentials)

    try:
        return _CLASS_CONVERSION_MAP[credentials_class](credentials)
    except KeyError as caught_exc:
        new_exc = ValueError(_CONVERT_ERROR_TMPL.format(credentials_class))
        raise new_exc from caught_exc



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to get a tuning job in Vertex AI."""

from googlecloudsdk.api_lib.ai.tuning_jobs import client
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.ai import constants
from googlecloudsdk.command_lib.ai import endpoint_util
from googlecloudsdk.command_lib.ai import validation
from googlecloudsdk.command_lib.ai.tuning_jobs import flags


@base.UniverseCompatible
@base.ReleaseTracks(
    base.ReleaseTrack.GA, base.ReleaseTrack.BETA, base.ReleaseTrack.ALPHA
)
class Describe(base.DescribeCommand):
  """Get detail information about the tuning job by given id.

  ## EXAMPLES

  To get a job ``123'' under project ``example'' in region
  ``us-central1'', run:

    $ {command} 123 --project=example --region=us-central1
  """

  @staticmethod
  def Args(parser):
    flags.AddTuningJobResourceArg(parser, "to describe")

  def Run(self, args):
    tuning_job_ref = args.CONCEPTS.tuning_job.Parse()
    region = tuning_job_ref.AsDict()["locationsId"]
    validation.ValidateRegion(
        region, available_regions=constants.SUPPORTED_TRAINING_REGIONS
    )

    version = (
        constants.GA_VERSION
        if self.ReleaseTrack() == base.ReleaseTrack.GA
        else constants.BETA_VERSION
    )
    with endpoint_util.AiplatformEndpointOverrides(
        version=version, region=region
    ):
      response = client.TuningJobsClient(version=version).Get(
          tuning_job_ref.RelativeName()
      )
      return response



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flag utils for clusters command group."""

import textwrap

from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import parser_arguments
from googlecloudsdk.command_lib.cluster_director.clusters import flag_types


def AddConfig(parser, api_version=None, required=False, hidden=False):
  """Adds a config flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(f"Unsupported API version for config: {api_version!r}")
  parser.add_argument(
      "--config",
      help="Configuration of the cluster specs in the form of a JSON object.",
      type=arg_parsers.ArgObject(
          spec=flag_types.FlagTypes(api_version).GetClusterFlagType(),
          enable_shorthand=True,
          disable_key_description=True,
          help_text=textwrap.dedent("""\
            See example below:

            --config='{
              "networkResources": {
                  "network0": {
                      "config": {
                          "newNetwork": {
                              "network": "network0"
                          }
                      }
                  }
              },
              "storageResources": {
                  "lustre0": {
                      "config": {
                          "newLustre": {
                              "capacityGb": 1200,
                              "filesystem": "lustre-fs"
                          }
                      }
                  }
              },
              "computeResources": {
                  "compute0": {
                      "config": {
                          "newOnDemandInstances": {
                              "machineType": "n2-standard-4",
                              "zone": "us-central1-a"
                          }
                      }
                  }
              },
              "orchestrator": {
                  "slurm": {
                      "nodeSets": [
                          {
                              "id": "nodeset0",
                              "computeId": "compute0"
                          }
                      ]
                  }
              }
          }' --format json
          """),
      ),
      required=required,
      hidden=hidden,
  )


def AddBlueprint(parser, api_version=None, hidden=False):
  """Adds a blueprint flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(f"Unsupported API version for blueprint: {api_version!r}")
  parser.add_argument(
      "--blueprint",
      help=textwrap.dedent("""
        Reference a pre-defined architecture (e.g., "a3-ultra").
      """),
      choices=[
          "a3-ultra",
          "a4-high-flex-start",
          "a4x-high",
          "g4",
          "h4d-highmem",
      ],
      hidden=hidden,
  )


def AddQuickstartCluster(parser, hidden=False):
  """Adds the --quickstart-cluster flag to the parser."""
  parser.add_argument(
      "--quickstart-cluster",
      action="store_true",
      help=(
          "If specified, creates a cluster with sensible defaults for compute,"
          " network, and orchestrator."
      ),
      hidden=hidden,
  )


def AddUpdateMask(parser, api_version=None, required=False, hidden=False):
  """Adds an update mask flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for update-mask: {api_version!r}"
    )
  parser.add_argument(
      "--update-mask",
      help=textwrap.dedent("""
        Update mask to specify the fields to update.

        For example --update-mask "description,labels"
      """),
      type=flag_types.UPDATE_MASK_OBJECT,
      required=required,
      hidden=hidden,
  )


def AddDescription(parser, api_version=None, hidden=False):
  """Adds a description flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for description: {api_version!r}"
    )
  parser.add_argument(
      "--description",
      help=textwrap.dedent("""
        Description of the cluster.

        For example --description "My cluster"
      """),
      type=str,
      hidden=hidden,
  )


def AddLabels(
    parser,
    name="labels",
    api_version=None,
    hidden=False,
    include_update_flags=False,
):
  """Adds a labels flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(f"Unsupported API version for labels: {api_version!r}")
  remove_flag_name = f"remove-{name}"
  if include_update_flags:
    name = f"add-{name}"
  parser.add_argument(
      f"--{name}",
      help=textwrap.dedent(f"""
        Cluster labels as key value pairs.

        For example --{name} key1=value1,key2=value2
      """),
      type=flag_types.LABEL,
      hidden=hidden,
  )
  if include_update_flags:
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Parameters to remove cluster label by key.

          For example --{remove_flag_name} key1,key2,...
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        hidden=hidden,
    )


def AddCreateNetwork(parser, api_version=None, hidden=False):
  """Adds a create network flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for create-network: {api_version!r}"
    )
  parser.add_argument(
      "--create-network",
      help=textwrap.dedent("""
        Parameters to create a VPC (Virtual Private Cloud) network.

        Required fields:
        - name
        Name: Must match the regex `[a-z]([-a-z0-9]*[a-z0-9])?`, be 1-63
        characters in length, and comply with RFC1035.

        Description: A description of the network. Maximum of 2048 characters.

        For example --create-network name=network-1,description="My network"
      """),
      type=flag_types.NETWORK_OBJECT,
      hidden=hidden,
  )


def AddNetworkSource(parser, api_version=None, required=False, hidden=False):
  """Adds a network flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(f"Unsupported API version for network: {api_version!r}")
  parser.add_argument(
      "--network",
      help=textwrap.dedent("""
        Reference of existing network name.
        If the network is in a different project (Shared VPC), specify
        the project ID using --network-project.

        For example --network network-1
      """),
      type=str,
      required=required,
      hidden=hidden,
  )


def AddNetworkProject(parser, api_version=None, hidden=False):
  """Adds a network project flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for network-project: {api_version!r}"
    )
  parser.add_argument(
      "--network-project",
      help=textwrap.dedent("""\
        Project ID of the project containing the network and subnetwork
        resources, if different from the cluster project (e.g. for Shared VPC).
      """),
      type=str,
      hidden=hidden,
  )


def AddSubnetSource(parser, api_version=None, required=False, hidden=False):
  """Adds a subnet flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(f"Unsupported API version for subnet: {api_version!r}")
  parser.add_argument(
      "--subnet",
      help=textwrap.dedent("""
        Reference of existing subnetwork name.
        If the subnetwork is in a different project (Shared VPC), specify
        the project ID using --network-project.

        For example --subnet regions/us-central1/subnetworks/subnet-1
      """),
      type=str,
      required=required,
      hidden=hidden,
  )


def AddCreateFilestores(
    parser,
    name="create-filestores",
    api_version=None,
    hidden=False,
    include_update_flags=False,
):
  """Adds a create filestores flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for create-filestores: {api_version!r}"
    )
  if include_update_flags:
    name = "add-new-filestore-instances"
  parser.add_argument(
      f"--{name}",
      help=textwrap.dedent("""
        Parameters to create a filestore instance.

        Required fields:
        - id
        - name
        - tier
        - capacityGb
        - fileshare

        For example --create-filestores id=my-fs,name=locations/us-central1/instances/filestore-1,tier=REGIONAL,capacityGb=1024,fileshare=share1

        id: Alphanumeric identifier of the filestore resource. Can be used to refer to this resource in storage-configs.
        name: Name of the filestore instance to create in your project in the format of locations/us-central1/instances/filestore-1.

        capacityGb: Size of the filestore in GiB. Must be between 1024 and 102400, and must meet scalability requirements described at
        https://cloud.google.com/filestore/docs/service-tiers.

        fileshare: The directory on a Filestore instance where all shared files
        are stored. Must match the regex `[a-z]([-a-z0-9]*[a-z0-9])?`, be 1-63
        characters in length, and comply with RFC1035.

        Supported protocol values:
        - NFSV3
        - NFSV41
        - If not specified, defaults to NFSV3

        Defaults:
        - protocol: NFSV3
      """),
      type=flag_types.FlagTypes(api_version).GetFilestoresObject(),
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )


def AddFilestores(
    parser,
    name="filestores",
    api_version=None,
    hidden=False,
    include_update_flags=False,
):
  """Adds a filestores flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(f"Unsupported API version for filestores: {api_version!r}")
  remove_flag_name = "remove-filestore-instances"
  if include_update_flags:
    name = "add-filestore-instances"
  parser.add_argument(
      f"--{name}",
      help=textwrap.dedent(f"""
        Reference of existing filestore instance.

        id: ID of the filestore resource, used to refer to this resource in storage-configs.
        name: Name of the existing filestore instance to import from your project in the format of locations/us-central1/instances/filestore-1.

        For example --{name} id=my-fs,name=locations/us-central1/instances/filestore-1
      """),
      type=flag_types.EXISTING_FILESTORES_TYPE,
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )
  if include_update_flags:
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Parameters to remove filestore instance config by filestore name.

          For example --{remove_flag_name} locations/us-central1/instances/filestore-1,locations/us-central1/instances/filestore-2,...
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        hidden=hidden,
    )


def AddCreateGcsBuckets(
    parser: parser_arguments.ArgumentInterceptor,
    name: str = "create-buckets",
    api_version: str = None,
    hidden: bool = False,
    include_update_flags: bool = False,
):
  """Adds a create Google Cloud Storage buckets flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for create-buckets: {api_version!r}"
    )
  if include_update_flags:
    name = "add-new-storage-buckets"
  alpha_help = textwrap.dedent(f"""
        Parameters to create a Google Cloud Storage bucket.

        Required fields:
        - id
        - name

        id: Alphanumeric identifier of the bucket resource. Can be used to refer to this resource in storage-configs.

        name: Name of the Cloud Storage bucket. Must be globally unique and omit any sensitive information.

        For example --{name} id=my-bucket,name=bucket-1,storageClass=STANDARD,terminalStorageClass=NEARLINE,enableHNS=true

        Supported storageClass values:
        - STANDARD
        - NEARLINE
        - COLDLINE
        - ARCHIVE

        Supported terminalStorageClass values:
        - NEARLINE
        - ARCHIVE

        Defaults:
        - storageClass: STANDARD

        Note:
        - Either storageClass or enableAutoclass can be set.
        - HNS: Hierarchical namespace. Hierarchical namespace buckets cannot use Autoclass.
      """)
  beta_help = textwrap.dedent(f"""
        Parameters to create a Google Cloud Storage bucket.

        Required fields:
        - id
        - name

        For example --{name} id=my-bucket,name=bucket-1,storageClass=STANDARD,enableHNS=true

        Supported storageClass values:
        - STANDARD
        - NEARLINE
        - COLDLINE
        - ARCHIVE

        Defaults:
        - storageClass: STANDARD

        Note:
        - Either storageClass or enableAutoclass can be set.
        - HNS: Hierarchical namespace. Hierarchical namespace buckets cannot use Autoclass.
      """)
  parser.add_argument(
      f"--{name}",
      help=alpha_help if api_version == "v1alpha" else beta_help,
      type=flag_types.FlagTypes(api_version).GetGcsBucketsObject(),
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )


def AddGcsBuckets(
    parser: parser_arguments.ArgumentInterceptor,
    name: str = "buckets",
    api_version: str = None,
    hidden: bool = False,
    include_update_flags: bool = False,
):
  """Adds a Google Cloud Storage buckets flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(f"Unsupported API version for buckets: {api_version!r}")
  remove_flag_name = f"remove-storage-{name}"
  if include_update_flags:
    name = f"add-storage-{name}"
  parser.add_argument(
      f"--{name}",
      help=textwrap.dedent(f"""
        Reference of existing Google Cloud Storage bucket.

        id: ID of the bucket resource, used to refer to this resource in storage-configs.
        name: Name of the existing Cloud Storage bucket to import from your project in the format of bucket-name.

        For example --{name} id=my-bucket,name=bucket-1
      """),
      type=flag_types.EXISTING_BUCKETS_TYPE,
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )
  if include_update_flags:
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Parameters to remove Google Cloud Storage bucket by bucket name.

          For example --{remove_flag_name} bucket-1,bucket-2,...
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        hidden=hidden,
    )


def AddCreateLustres(
    parser: parser_arguments.ArgumentInterceptor,
    name: str = "create-lustres",
    api_version: str = None,
    hidden: bool = False,
    include_update_flags: bool = False,
):
  """Adds a create lustres flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for create-lustres: {api_version!r}"
    )
  if include_update_flags:
    name = "add-new-lustre-instances"
  parser.add_argument(
      f"--{name}",
      help=textwrap.dedent(f"""
        Parameters to create a Lustre instance.

        Required fields:
        - id
        - name
        - capacityGb
        - filesystem.

        id: ID of the lustre resource, used to refer to this resource in storage-configs.

        name: Name of the Managed Lustre instance to create in your project in the format of locations/us-central1/instances/lustre-1.

        For example --{name} id=my-lustre,name=locations/us-central1/instances/lustre-1,capacityGb=1024,filesystem=fs-1,perUnitStorageThroughput=1000

        Values for perUnitStorageThroughput: 125, 250, 500, 1000
      """),
      type=flag_types.LUSTRES_OBJECT,
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )


def AddLustres(
    parser,
    name="lustres",
    api_version=None,
    hidden=False,
    include_update_flags=False,
):
  """Adds a lustres flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(f"Unsupported API version for lustres: {api_version!r}")
  remove_flag_name = "remove-lustre-instances"
  if include_update_flags:
    name = "add-lustre-instances"
  parser.add_argument(
      f"--{name}",
      help=textwrap.dedent(f"""
        Reference of existing Lustre instance.

        id: ID of the lustre resource, used to refer to this resource in storage-configs.
        name: Name of the existing Managed Lustre instance to import from your project in the format of locations/us-central1/instances/lustre-1.

        For example --{name} id=my-lustre,name=locations/us-central1/instances/lustre-1
      """),
      type=flag_types.EXISTING_LUSTRES_TYPE,
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )
  if include_update_flags:
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Parameters to remove lustre instance config by lustre name.

          For example --{remove_flag_name} locations/us-central1/instances/lustre-1,locations/us-central1/instances/lustre-2,...
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        hidden=hidden,
    )


def AddOnDemandInstances(
    parser,
    name="on-demand-instances",
    api_version=None,
    hidden=False,
    include_update_flags=False,
):
  """Adds an on-demand instances flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for on-demand-instances: {api_version!r}"
    )
  remove_flag_name = f"remove-{name}"
  if include_update_flags:
    name = f"add-{name}"
  alpha_help = textwrap.dedent(f"""
        Parameters to define cluster on demand instances.

        For example --{name} id=c1,zone=us-central1-a,machineType=n1-standard-1
      """)
  beta_help = textwrap.dedent(f"""
        Parameters to define cluster on demand instances.

        For example --{name} id=c1,zone=us-central1-a,machineType=n1-standard-1
      """)
  parser.add_argument(
      f"--{name}",
      help=alpha_help if api_version == "v1alpha" else beta_help,
      type=flag_types.FlagTypes(api_version).GetOnDemandInstancesObject(),
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )
  if include_update_flags:
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Parameters to remove on demand instances config by compute id.

          For example --{remove_flag_name} c1,c2,...
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        hidden=hidden,
    )


def AddSpotInstances(
    parser,
    name="spot-instances",
    api_version=None,
    hidden=False,
    include_update_flags=False,
):
  """Adds a spot VM instances flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for spot-instances: {api_version!r}"
    )
  remove_flag_name = f"remove-{name}"
  if include_update_flags:
    name = f"add-{name}"
  alpha_help = textwrap.dedent(f"""
        Parameters to define cluster spot instances.

        For example --{name} id=c1,zone=us-central1-a,machineType=n1-standard-1
      """)
  beta_help = textwrap.dedent(f"""
        Parameters to define cluster spot instances.

        For example --{name} id=c1,zone=us-central1-a,machineType=n1-standard-1
      """)
  parser.add_argument(
      f"--{name}",
      help=alpha_help if api_version == "v1alpha" else beta_help,
      type=flag_types.FlagTypes(api_version).GetSpotInstancesObject(),
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )
  if include_update_flags:
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Parameters to remove spot instance config by compute id.

          For example --{remove_flag_name} c1,c2,...
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        hidden=hidden,
    )


def AddReservedInstances(
    parser,
    name="reserved-instances",
    api_version=None,
    hidden=False,
    include_update_flags=False,
):
  """Adds a reserved instances flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for reserved-instances: {api_version!r}"
    )
  remove_flag_name = f"remove-{name}"
  if include_update_flags:
    name = f"add-{name}"
  alpha_help = textwrap.dedent(f"""
        Parameters to define cluster reserved instances.

        For example --{name} id=c1,reservation=zones/us-central1-a/reservations/reservation-1

        Exactly one of reservation, reservation-block, or reservation-sub-block must be provided.
        reservation: The name of the reservation to use, in the format zones/us-central1-a/reservations/reservation-1.
        reservationBlock: The name of the reservation block to use, in the format zones/us-central1-a/reservations/reservation-1/reservationBlocks/block-1.
        reservationSubBlock: The name of the reservation sub-block to use, in the format zones/us-central1-a/reservations/reservation-1/reservationBlocks/block-1/reservationSubBlocks/sub-block-1.
      """)
  beta_help = textwrap.dedent(f"""
        Parameters to define cluster reserved instances.

        For example --{name} id=c1,reservation=zones/us-central1-a/reservations/reservation-1
      """)
  parser.add_argument(
      f"--{name}",
      help=alpha_help if api_version == "v1alpha" else beta_help,
      type=flag_types.FlagTypes(api_version).GetReservedInstancesObject(),
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )
  if include_update_flags:
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Parameters to remove reserved instance config by compute id.

          For example --{remove_flag_name} c1,c2,...
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        hidden=hidden,
    )


def AddFlexStartInstances(
    parser,
    name="flex-start-instances",
    api_version=None,
    hidden=False,
    include_update_flags=False,
):
  """Adds a flex-start VM instances flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for flex-start-instances: {api_version!r}"
    )
  remove_flag_name = f"remove-{name}"
  if include_update_flags:
    name = f"add-{name}"
  alpha_help = textwrap.dedent(f"""
        Parameters to define cluster Flex Start instances.

        For example --{name} id=c1,zone=us-central1-a,machineType=n1-standard-1,maxDuration=10000s
      """)
  beta_help = textwrap.dedent(f"""
        Parameters to define cluster Flex Start instances.

        For example --{name} id=c1,zone=us-central1-a,machineType=n1-standard-1,maxDuration=10000s
      """)
  parser.add_argument(
      f"--{name}",
      help=alpha_help if api_version == "v1alpha" else beta_help,
      type=flag_types.FlagTypes(api_version).GetFlexStartInstancesObject(),
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )
  if include_update_flags:
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Parameters to remove Flex Start instance config by compute id.

          For example --{remove_flag_name} c1,c2,...
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        hidden=hidden,
    )


def AddSlurmNodeSets(
    parser,
    name="slurm-node-sets",
    api_version=None,
    required=False,
    hidden=False,
    include_update_flags=False,
):
  """Adds a slurm nodesets flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for slurm-node-sets: {api_version!r}"
    )
  update_flag_name = f"update-{name}"
  remove_flag_name = f"remove-{name}"
  if include_update_flags:
    name = f"add-{name}"
  alpha_help = textwrap.dedent(f"""
        Parameters to define slurm cluster nodeset config.

        Required fields:
        - id

        For example --{name} id=ns1,computeId=c1,type=gce,staticNodeCount=1,maxDynamicNodeCount=2,startupScript="echo hello",bootDisk=type=hyperdisk-balanced,storagePools=[zones/us-central1-a/storagePools/pool1]

        To configure a nodeset backed by Google Kubernetes Engine, use type=gke. If type=gke is
        specified, Compute Engine specific fields (labels, startupScript, bootDisk,
        startupScriptTimeout) cannot be used, but container-specific fields
        (container-resource-labels, container-startup-script) may be used.
        For example --{name} id=ns1,computeId=c1,type=gke
        For example --{name} id=ns1,computeId=c1,type=gke,container-resource-labels="k1=v1",container-startup-script="echo hello"

        Defaults:
        - type: gce

        Note:
        - startupScript:
          - Either str or file_path
          - For file_path, only bash file format (.sh or .bash) is supported.
          - For file_path, only absolute path is supported.
        - bootDisk.storagePools:
          - At most 1 storage pool is supported.
          - storagePools: The name of the storage pool to use, in the format zones/us-central1-a/storagePools/pool1.
      """)
  beta_help = textwrap.dedent(f"""
        Parameters to define slurm cluster nodeset config.

        Required fields:
        - id

        For example --{name} id=ns1,computeId=c1,staticNodeCount=1,maxDynamicNodeCount=2,computeInstance=[startupScript="echo hello",labels="k1=v1"]

        Defaults:

        Note:
        - startupScript:
          - Either str or file_path
          - For file_path, only bash file format (.sh or .bash) is supported.
          - For file_path, only absolute path is supported.
      """)
  parser.add_argument(
      f"--{name}",
      help=alpha_help if api_version == "v1alpha" else beta_help,
      type=flag_types.FlagTypes(api_version).GetSlurmNodeSetsObject(),
      action=arg_parsers.FlattenAction(),
      required=required,
      hidden=hidden,
  )
  if include_update_flags:
    alpha_update_help = textwrap.dedent(f"""
          Parameters to define and update slurm cluster nodeset config.

          For example --{update_flag_name} id=ns1,staticNodeCount=1,maxDynamicNodeCount=2

          To update a node set backed by GKE, use container-resource-labels or container-startup-script.
          For example --{update_flag_name} id=ns1,type=gke,container-resource-labels="k1=v1",container-startup-script="echo hello"
        """)
    beta_update_help = textwrap.dedent(f"""
          Parameters to define and update slurm cluster nodeset config.

          For example --{update_flag_name} id=ns1,staticNodeCount=1,maxDynamicNodeCount=2,computeInstance=[startupScript="echo hello"]
        """)
    parser.add_argument(
        f"--{update_flag_name}",
        help=alpha_update_help
        if api_version == "v1alpha"
        else beta_update_help,
        type=flag_types.FlagTypes(api_version).GetSlurmNodeSetsObject(),
        action=arg_parsers.FlattenAction(),
        required=required,
        hidden=hidden,
    )
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Parameters to remove slurm nodeset config by nodeset id.

          For example --{remove_flag_name} ns1,ns2,...
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        required=required,
        hidden=hidden,
    )


def AddSlurmPartitions(
    parser,
    name="slurm-partitions",
    api_version=None,
    required=False,
    hidden=False,
    include_update_flags=False,
):
  """Adds a slurm partitions flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for slurm-partitions: {api_version!r}"
    )
  update_flag_name = f"update-{name}"
  remove_flag_name = f"remove-{name}"
  if include_update_flags:
    name = f"add-{name}"
  alpha_help = textwrap.dedent(f"""
        Parameters to define slurm cluster partitions.

        For example --{name} id=p1,nodesetIds=[ns1,ns2]
      """)
  beta_help = textwrap.dedent(f"""
        Parameters to define slurm cluster partitions.

        For example --{name} id=p1,nodesetIds=[ns1,ns2]
      """)
  parser.add_argument(
      f"--{name}",
      help=alpha_help if api_version == "v1alpha" else beta_help,
      type=flag_types.FlagTypes(api_version).GetSlurmPartitionsObject(),
      action=arg_parsers.FlattenAction(),
      required=required,
      hidden=hidden,
  )
  if include_update_flags:
    alpha_update_help = textwrap.dedent(f"""
          Parameters to define and update slurm cluster partition config.

          For example --{update_flag_name} id=p1,nodesetIds=[ns1,ns2]
        """)
    beta_update_help = textwrap.dedent(f"""
          Parameters to define and update slurm cluster partition config.

          For example --{update_flag_name} id=p1,nodesetIds=[ns1,ns2]
        """)
    parser.add_argument(
        f"--{update_flag_name}",
        help=alpha_update_help
        if api_version == "v1alpha"
        else beta_update_help,
        type=flag_types.FlagTypes(api_version).GetSlurmPartitionsUpdateObject(),
        action=arg_parsers.FlattenAction(),
        required=required,
        hidden=hidden,
    )
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Parameters to remove slurm partition config by partition id.

          For example --{remove_flag_name} p1,p2,...
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        required=required,
        hidden=hidden,
    )


def AddSlurmDefaultPartition(parser, api_version=None, hidden=False):
  """Adds a slurm default partition flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for slurm-default-partition: {api_version!r}"
    )
  parser.add_argument(
      "--slurm-default-partition",
      help=textwrap.dedent("""
        Parameters to define slurm cluster default partition.

        For example --slurm-default-partition p1
      """),
      type=str,
      hidden=hidden,
  )


def AddSlurmLoginNode(
    parser,
    name="slurm-login-node",
    api_version=None,
    required=False,
    hidden=False,
    include_update_flags=False,
):
  """Adds a slurm login node flag for the given API version."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for slurm-login-node: {api_version!r}"
    )
  ft = flag_types.FlagTypes(api_version)
  alpha_create_help = textwrap.dedent("""
        Parameters to define slurm cluster login node.

        Required:
        - machineType
        - zone

        For example --slurm-login-node machineType=n1-standard-1,zone=us-central1-a,count=1

        Defaults:
        - count: 1
        - enableOsLogin: true
        - enablePublicIPs: true
        - bootDisk.sizeGb: 100

        Note:
        - startupScript:
          - Either str or file_path
          - For file_path, only bash file format (.sh or .bash) is supported.
          - For file_path, only absolute path is supported.
        - bootDisk:
          - Must be 50 GB or greater.
        - bootDisk.storagePools:
          - At most 1 storage pool is supported.
          - storagePools: The name of the storage pool to use, in the format zones/us-central1-a/storagePools/pool1.
      """)
  beta_create_help = textwrap.dedent("""
        Parameters to define slurm cluster login node.

        Required:
        - machineType
        - zone

        For example --slurm-login-node machineType=n1-standard-1,zone=us-central1-a,count=1

        Defaults:
        - count: 1
        - enableOsLogin: true
        - enablePublicIPs: true
        - bootDisk.sizeGb: 100

        Note:
        - startupScript:
          - Either str or file_path
          - For file_path, only bash file format (.sh or .bash) is supported.
          - For file_path, only absolute path is supported.
        - bootDisk:
          - Must be 50 GB or greater.
      """)
  flag_name = name
  if include_update_flags:
    flag_name = f"update-{name}"
    alpha_update_help = textwrap.dedent(f"""
        Parameters to update slurm cluster login node.
        Only bootDisk, count and startupScript can be updated.

        For example --{flag_name} count=2,startupScript="echo hello",bootDisk=type=hyperdisk-balanced,storagePools=[zones/us-central1-a/storagePools/pool1]

        Note:
        - bootDisk.storagePools:
          - At most 1 storage pool is supported.
          - storagePools: The name of the storage pool to use, in the format zones/us-central1-a/storagePools/pool1.
    """)
    beta_update_help = textwrap.dedent(f"""
        Parameters to update slurm cluster login node.
        Only bootDisk, count and startupScript can be updated.

        For example --{flag_name} count=2,startupScript="echo hello"
    """)
    help_text = (
        alpha_update_help if api_version == "v1alpha" else beta_update_help
    )
    flag_type = ft.GetSlurmLoginNodeUpdateObject()
  else:
    help_text = (
        alpha_create_help if api_version == "v1alpha" else beta_create_help
    )
    flag_type = ft.GetSlurmLoginNodeObject()

  parser.add_argument(
      f"--{flag_name}",
      help=help_text,
      type=flag_type,
      required=required,
      hidden=hidden,
  )


def _AddScriptFlags(
    parser, name, help_kind, api_version, hidden, include_update_flags
):
  """Helper to add script flags."""
  if api_version not in ["v1alpha", "v1beta", "v1"]:
    raise ValueError(
        f"Unsupported API version for {help_kind}: {api_version!r}"
    )
  remove_flag_name = f"remove-{name}"
  if include_update_flags:
    flag_name = f"add-{name}"
  else:
    flag_name = name
  parser.add_argument(
      f"--{flag_name}",
      help=textwrap.dedent(f"""
        {help_kind}.

        For example --{flag_name} script1.sh,script2.sh
      """),
      type=arg_parsers.ArgList(element_type=str),
      action=arg_parsers.FlattenAction(),
      hidden=hidden,
  )
  if include_update_flags:
    parser.add_argument(
        f"--{remove_flag_name}",
        help=textwrap.dedent(f"""
          Scripts to remove from {help_kind}.

          For example --{remove_flag_name} script1.sh,script2.sh
        """),
        type=arg_parsers.ArgList(element_type=str),
        action=arg_parsers.FlattenAction(),
        hidden=hidden,
    )


def AddSlurmPrologBashScripts(
    parser, api_version=None, hidden=False, include_update_flags=False
):
  """Adds a slurm prolog bash scripts flag for the given API version."""
  _AddScriptFlags(
      parser,
      "slurm-prolog-scripts",
      "Slurm prolog bash scripts",
      api_version,
      hidden,
      include_update_flags,
  )


def AddSlurmEpilogBashScripts(
    parser, api_version=None, hidden=False, include_update_flags=False
):
  """Adds a slurm epilog bash scripts flag for the given API version."""
  _AddScriptFlags(
      parser,
      "slurm-epilog-scripts",
      "Slurm epilog bash scripts",
      api_version,
      hidden,
      include_update_flags,
  )


def AddSlurmTaskPrologBashScripts(
    parser, api_version=None, hidden=False, include_update_flags=False
):
  """Adds a slurm task prolog bash scripts flag for the given API version.

  Args:
    parser: The argparse parser.
    api_version: The API version to use (e.g., "v1alpha").
    hidden: Whether the flag should be hidden.
    include_update_flags: Whether to include flags for update commands.

  Raises:
    ValueError: If the api_version is not supported.
  """
  if api_version not in ["v1alpha"]:
    raise ValueError(
        "Unsupported API version for slurm-task-prolog-scripts:"
        f" {api_version!r}"
    )
  _AddScriptFlags(
      parser,
      "slurm-task-prolog-scripts",
      "Slurm task prolog bash scripts",
      api_version,
      hidden,
      include_update_flags,
  )


def AddSlurmTaskEpilogBashScripts(
    parser, api_version=None, hidden=False, include_update_flags=False
):
  """Adds a slurm task epilog bash scripts flag for the given API version.

  Args:
    parser: The argparse parser.
    api_version: The API version to use (e.g., "v1alpha").
    hidden: Whether the flag should be hidden.
    include_update_flags: Whether to include flags for update commands.

  Raises:
    ValueError: If the api_version is not supported.
  """
  if api_version not in ["v1alpha"]:
    raise ValueError(
        "Unsupported API version for slurm-task-epilog-scripts:"
        f" {api_version!r}"
    )
  _AddScriptFlags(
      parser,
      "slurm-task-epilog-scripts",
      "Slurm task epilog bash scripts",
      api_version,
      hidden,
      include_update_flags,
  )


def AddSlurmConfig(
    parser, api_version=None, hidden=False, include_update_flags=False
):
  """Adds a slurm config flag for the given API version."""
  if api_version not in ["v1alpha"]:
    raise ValueError(
        f"Unsupported API version for slurm-config: {api_version!r}"
    )
  flag_name = "slurm-config"
  if include_update_flags:
    flag_name = f"update-{flag_name}"
  parser.add_argument(
      f"--{flag_name}",
      help=textwrap.dedent(f"""
        Parameters to define slurm cluster config.

        For example --{flag_name} healthCheckInterval=10,healthCheckNodeState=IDLE,healthCheckProgram=/usr/bin/true
      """),
      type=flag_types.SLURM_CONFIG_TYPE,
      hidden=hidden,
  )


def AddSlurmDisableHealthCheckProgram(
    parser, api_version=None, hidden=False, include_update_flags=False
):
  """Adds a slurm disable health check program flag for the given API version."""
  if api_version not in ["v1alpha"]:
    raise ValueError(
        "Unsupported API version for slurm-disable-health-check-program:"
        f" {api_version!r}"
    )
  flag_name = "slurm-disable-health-check-program"
  if include_update_flags:
    flag_name = f"update-{flag_name}"
  parser.add_argument(
      f"--{flag_name}",
      help=textwrap.dedent(f"""
        If true, health checking is disabled, and health_check_interval,
        health_check_node_state, and health_check_program should not be passed in.

        For example --{flag_name}
      """),
      action="store_true",
      hidden=hidden,
  )



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Update the configuration of an insight config."""

import datetime

from googlecloudsdk.api_lib.developer_connect.insights_configs import insights_config
from googlecloudsdk.api_lib.util import exceptions
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.developer_connect import flags
from googlecloudsdk.command_lib.developer_connect import resource_args
from googlecloudsdk.core import log

DETAILED_HELP = {
    'DESCRIPTION': """
          Create an insights config.
          """,
    'EXAMPLES': """
          To create an insights config with an apphub application, run:

            $ {command} insights-config-name --app-hub-application=projects/my-project/locations/us-central1/applications/my-app-hub-application

          To create an insights config with projects, run:

            $ {command} insights-config-name --target-projects=project1,project2
          """,
}


@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA, base.ReleaseTrack.GA
)
@base.DefaultUniverseOnly
class Create(base.CreateCommand):
  """Create an insight config."""

  @classmethod
  def Args(cls, parser):
    """Adds arguments for this command."""
    try:
      resource_args.AddInsightConfigResourceArg(parser, verb='create')
    except exceptions.HttpException as e:
      log.status.Print('Failed to add insight config resource argument.')
      raise e

    # Use cls.ReleaseTrack() to check if we are in Alpha
    is_alpha = cls.ReleaseTrack() == base.ReleaseTrack.ALPHA

    source_group = parser.add_mutually_exclusive_group(required=not is_alpha)
    flags.AddAppHubApplicationArgument(source_group)
    flags.AddTargetProjectsArgument(source_group)

    flags.AddArtifactConfigsArgument(parser)
    flags.AddSourceConfigArgument(parser)

  def Run(self, args):
    max_wait = datetime.timedelta(seconds=30)
    client = insights_config.InsightsConfigClient(self.ReleaseTrack())
    insights_config_ref = args.CONCEPTS.insights_config.Parse()
    try:
      operation = client.create(
          insight_config_ref=insights_config_ref,
          app_hub=args.app_hub_application,
          target_projects=args.target_projects,
          user_artifact_configs=args.artifact_configs,
          source_config=args.source_config,
      )
    except exceptions.HttpException as e:
      log.status.Print('Failed to create the insight config {}.'.format(
          insights_config_ref.RelativeName()
      ))
      raise e

    log.status.Print('Creating the insight config {}.'.format(
        insights_config_ref.RelativeName()
    ))

    return client.wait_for_operation(
        operation_ref=client.get_operation_ref(operation),
        message='Waiting for operation [{}] to be completed...'
        .format(
            client.get_operation_ref(operation).RelativeName()),
        has_result=True,
        max_wait=max_wait,
    )
Create.detailed_help = DETAILED_HELP



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Factory for ExecutionConfig message."""


from apitools.base.py import encoding
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.command_lib.dataproc.shared_messages import (
    authentication_config_factory as acf)
import six


class ExecutionConfigFactory(object):
  """Factory for ExecutionConfig message.

  Add ExecutionConfig related arguments to argument parser and create
  ExecutionConfig message from parsed arguments.
  """

  def __init__(self, dataproc, auth_config_factory_override=None):
    """Factory class for ExecutionConfig message.

    Args:
      dataproc: A api_lib.dataproc.Dataproc instance.
      auth_config_factory_override: Override the default
      AuthenticationConfigFactory instance. This is a keyword argument.
    """
    self.dataproc = dataproc

    self.auth_config_factory = auth_config_factory_override
    if not self.auth_config_factory:
      self.auth_config_factory = (
          acf.AuthenticationConfigFactory(
              self.dataproc
          )
      )

  def GetMessage(self, args):
    """Builds an ExecutionConfig instance.

    Build a ExecutionConfig instance according to user settings.
    Returns None if all fileds are None.

    Args:
      args: Parsed arguments.

    Returns:
      ExecutionConfig: A ExecutionConfig instance. None if all fields are
      None.
    """
    kwargs = {}

    if args.tags:
      kwargs['networkTags'] = args.tags

    resource_manager_tags = args.resource_manager_tag
    if resource_manager_tags:
      merged_tags = {}
      for d in resource_manager_tags:
        merged_tags.update(d)
      kwargs['resourceManagerTags'] = encoding.DictToAdditionalPropertyMessage(
          merged_tags,
          self.dataproc.messages.ExecutionConfig.ResourceManagerTagsValue,
          sort_items=True,
      )

    if args.network:
      kwargs['networkUri'] = args.network

    if args.subnet:
      kwargs['subnetworkUri'] = args.subnet

    if args.service_account:
      kwargs['serviceAccount'] = args.service_account

    if args.kms_key:
      kwargs['kmsKey'] = args.kms_key

    if hasattr(args, 'max_idle') and args.max_idle:
      # ExecutionConfig message expects duration in seconds
      kwargs['idleTtl'] = six.text_type(args.max_idle) + 's'

    if args.ttl:
      # ExecutionConfig message expects duration in seconds
      kwargs['ttl'] = six.text_type(args.ttl) + 's'

    if args.staging_bucket:
      kwargs['stagingBucket'] = args.staging_bucket

    authentication_config = (
        self.auth_config_factory.GetMessage(args)
    )
    if authentication_config:
      kwargs['authenticationConfig'] = authentication_config

    if not kwargs:
      return None

    return self.dataproc.messages.ExecutionConfig(**kwargs)


def AddArguments(parser):
  """Adds ExecutionConfig related arguments to parser."""

  parser.add_argument(
      '--service-account',
      help='The IAM service account to be used for a batch/session job.')

  network_group = parser.add_mutually_exclusive_group()
  network_group.add_argument(
      '--network', help='Network URI to connect network to.')
  network_group.add_argument(
      '--subnet',
      help=('Subnetwork URI to connect network to. Subnet must have Private '
            'Google Access enabled.'))

  parser.add_argument(
      '--tags',
      type=arg_parsers.ArgList(),
      metavar='TAGS',
      default=[],
      help='Network tags for traffic control.')

  parser.add_argument('--kms-key', help='Cloud KMS key to use for encryption.')

  parser.add_argument(
      '--resource-manager-tag',
      type=arg_parsers.ArgDict(min_length=1, max_length=1),
      action='append',
      default=[],
      metavar='KEY=VALUE',
      help=(
          'Resource Manager Tags to be associated with the compute resources'
          ' created for the workload. Only one key-value pair can be specified'
          ' per flag. Repeat the flag to specify multiple tags.'
      ),
  )

  parser.add_argument(
      '--staging-bucket',
      help="""\
      The Cloud Storage bucket to use to store job dependencies, config files,
      and job driver console output. If not specified, the default [staging bucket]
      (https://cloud.google.com/dataproc-serverless/docs/concepts/buckets) is used.
      """,
  )

  parser.add_argument(
      '--ttl',
      type=arg_parsers.Duration(),
      help="""
      The duration after the workload will be unconditionally terminated,
      for example, '20m' or '1h'. Run
      [gcloud topic datetimes](https://cloud.google.com/sdk/gcloud/reference/topic/datetimes)
      for information on duration formats.""",
  )

  acf.AddArguments(parser)



# -*- coding: utf-8 -*- #
# Copyright 2024 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A library used to support Managed Service for Apache Kafka commands."""

import re

from apitools.base.py import encoding
from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk import core
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.calliope import actions
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import properties

# Retrieve all message type for conversions from gcloud primitives to
# apitool types.
_MESSAGE = apis.GetMessagesModule("managedkafka", "v1")
SASL_PORT = "9092"
MTLS_PORT = "9192"
CONTEXTS_RESOURCE_PATH = "/contexts/"
SUBJECTS_RESOURCE_PATH = "/subjects/"
SUBJECTS_MODE_RESOURCE_PATH = "/mode/"
SUBJECTS_CONFIG_RESOURCE_PATH = "/config/"


def ValidateCPU(cpu):
  """Validate CPU >= 3."""
  if cpu < 3:
    raise exceptions.BadArgumentException("--cpu", "CPU must be at least 3")
  return cpu


def PrepareUpdateWithSubnets(_, args, request):
  """Prepare the update request with the information from the subnet flag.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
    The updated request with the subnet.
  """
  if not args.subnets:
    return request

  # The cluster is not created yet if only the subnet flag is set. This is
  # because we don't formally map the subnet to the request the same way we do
  # for other flags. Instead, subnets require special handling with the use of
  # hooks.
  if not request.cluster:
    request.cluster = {}

  subnet_update_mask = "gcpConfig.accessConfig.networkConfigs"
  request.updateMask = AppendUpdateMask(request.updateMask, subnet_update_mask)
  return MapSubnetsToNetworkConfig(_, args, request)


def PrepareUpdateWithCaPools(_, args, request):
  """Prepare the update request with the information from the mTLS CA pool flag.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
    The updated request with the CA pool.
  """

  # If the user is clearing the CA pools, we need to add the update mask.
  # This flag is guarded by a mutex and will not conflict with the ca pool flag.
  if args.clear_mtls_ca_pools:
    request.updateMask = AppendUpdateMask(
        request.updateMask, "tlsConfig.trustConfig.casConfigs"
    )
    return request

  # If the there are no CA pools to update, return the request as is.
  if not args.mtls_ca_pools:
    return request

  # The cluster is not created yet if only the CA pool flag is set. This is
  # because we don't formally map the CA pool to the request the same way we do
  # for other flags. Instead, CA pools require special handling with the use of
  # hooks.
  if not request.cluster:
    request.cluster = {}

  ca_pool_update_mask = "tlsConfig.trustConfig.casConfigs"
  request.updateMask = AppendUpdateMask(request.updateMask, ca_pool_update_mask)
  return MapCaPoolsToCASConfig(_, args, request)


def PrepareUpdateWithPublicCluster(_, args, request):
  """Prepare the update request with the public cluster flag.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
    The updated request with public cluster config.
  """
  public_cluster = getattr(args, "public_cluster", None)
  allowed_ranges = getattr(args, "allowed_source_ip_ranges", None)

  if public_cluster is None and not allowed_ranges:
    return request

  if not request.cluster:
    request.cluster = {}

  if not request.cluster.gcpConfig:
    request.cluster.gcpConfig = {}

  # --no-public-cluster
  if public_cluster is not None and not public_cluster:
    if allowed_ranges:
      raise exceptions.InvalidArgumentException(
          "--allowed-source-ip-ranges",
          "Cannot specify allowed source IP ranges when disabling public"
          " cluster.",
      )
    # Clear publicClusterConfig
    request.updateMask = AppendUpdateMask(
        request.updateMask, "gcpConfig.accessConfig.publicClusterConfig"
    )
    # We don't populate it in request.cluster, so it will be cleared
    return request

  if public_cluster and not allowed_ranges:
    raise exceptions.InvalidArgumentException(
        "--allowed-source-ip-ranges",
        "Must specify at least one IPv4 CIDR range when public cluster is"
        " enabled.",
    )

  # If they specify allowed_source_ip_ranges, they are updating or enabling it
  if allowed_ranges:
    request.updateMask = AppendUpdateMask(
        request.updateMask, "gcpConfig.accessConfig.publicClusterConfig"
    )
    if not request.cluster.gcpConfig.accessConfig:
      request.cluster.gcpConfig.accessConfig = {}

    public_config = {"allowedSourceIpRanges": allowed_ranges}
    request.cluster.gcpConfig.accessConfig.publicClusterConfig = (
        encoding.DictToMessage(public_config, _MESSAGE.PublicClusterConfig)
    )

  return request


def AppendUpdateMask(update_mask, new_mask):
  """Handles appending a new mask to an existing mask.

  Args:
    update_mask: the existing update mask.
    new_mask: the new mask to append.

  Returns:
    The fully appended update mask.
  """
  update_mask = f"{update_mask},{new_mask}"
  return update_mask if update_mask[0] != "," else update_mask[1:]


def MapSubnetsToNetworkConfig(_, args, request):
  """Maps the list of subnets from the flag to the API fields in the request.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
    The updated request with networkConfig in the JSON format.
  """
  # Reference the existing GCP config if already created for the request.
  if not request.cluster.gcpConfig:
    request.cluster.gcpConfig = {}
  request.cluster.gcpConfig.accessConfig = {"networkConfigs": []}
  for subnet in args.subnets:
    network_config = {"subnet": subnet}
    request.cluster.gcpConfig.accessConfig.networkConfigs.append(
        encoding.DictToMessage(network_config, _MESSAGE.NetworkConfig)
    )

  # Handle public cluster config mapping
  public_cluster = getattr(args, "public_cluster", None)
  allowed_ranges = getattr(args, "allowed_source_ip_ranges", None)

  if public_cluster:
    if not allowed_ranges:
      raise exceptions.InvalidArgumentException(
          "--allowed-source-ip-ranges",
          "Must specify at least one IPv4 CIDR range when public cluster is"
          " enabled.",
      )
    public_config = {"allowedSourceIpRanges": allowed_ranges}
    request.cluster.gcpConfig.accessConfig.publicClusterConfig = (
        encoding.DictToMessage(public_config, _MESSAGE.PublicClusterConfig)
    )
  elif allowed_ranges:
    raise exceptions.InvalidArgumentException(
        "--public-cluster",
        "Must enable public cluster to specify allowed source IP ranges.",
    )

  return request


def MapCaPoolsToCASConfig(_, args, request):
  """Maps the list of CA pools from the flag to the API fields in the request.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
    The updated request with CertificateAuthorityServiceConfig in the JSON
    format.
  """
  if not args.mtls_ca_pools:
    return request

  # Reference the existing CAS config if already created for the request.
  if not request.cluster.tlsConfig:
    request.cluster.tlsConfig = {}
  request.cluster.tlsConfig.trustConfig = {"casConfigs": []}
  for ca_pool in args.mtls_ca_pools:
    cas_config = {"caPool": ca_pool}
    request.cluster.tlsConfig.trustConfig.casConfigs.append(
        encoding.DictToMessage(
            cas_config, _MESSAGE.CertificateAuthorityServiceConfig
        )
    )
  return request


def ListWithBootstrapAddr(response, _):
  """Synthesizes the bootstrap address to the response for a list request.

  Args:
    response: the payload to return.
    _: list of flags.

  Returns:
    The updated clusters with the bootstrap.
  """
  return [
      SynthesizeBootstrapAddr(cluster, cluster.name) for cluster in response
  ]


def DescribeWithBootstrapAddr(response, _):
  """Synthesizes the bootstrap address to the response for a describe request.

  Args:
    response: the payload to return.
    _: list of flags.

  Returns:
    The updated cluster with the bootstrap.
  """
  return SynthesizeBootstrapAddr(response, response.name)


def SynthesizeBootstrapAddr(response, cluster):
  """Synthesizes the bootstrap address to the response.

  Args:
    response: the payload to update.
    cluster: the fully qualifed name of the cluster.

  Returns:
    The updated cluster with the bootstrap
  """
  # The fully qualified name will always be consistent. We also have to use the
  # fully qualifed name instead of the resource directly to support both
  # `describe` and `list`.
  name = cluster.split("/")[5]
  location = cluster.split("/")[3]
  project = core.properties.VALUES.core.project.Get()
  domain_prefixed_project = project.split(":")
  if len(domain_prefixed_project) == 2:
    project = f"{domain_prefixed_project[1]}.{domain_prefixed_project[0]}"
  bootstrap = f"bootstrap.{name}.{location}.managedkafka.{project}.cloud.goog"
  synthesized = core.resource.resource_projector.MakeSerializable(response)
  synthesized["bootstrapAddress"] = f"{bootstrap}:{SASL_PORT}"
  if hasattr(response, "tlsConfig") and response.tlsConfig:
    synthesized["bootstrapAddressMTLS"] = f"{bootstrap}:{MTLS_PORT}"
  return synthesized


def UpdateTopics(_, args, request):
  """Load the topics JSON from the argument to the request.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
    The updated request with topics.
  """
  topics = core.yaml.load(args.topics_file)
  request.consumerGroup = {
      "topics": encoding.DictToMessage(
          topics, _MESSAGE.ConsumerGroup.TopicsValue
      )
  }
  request.updateMask = "topics"
  return request


def MapConnectParamsToNetworkConfig(_, args, request):
  """Maps subnets and DNS names to the network config API field.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
    The updated request with networkConfig in the JSON format.
  """
  # If no network config flags are provided (such as in the case of an update),
  # we don't need to create a network config.
  if not args.primary_subnet and not args.dns_name:
    return request
  # Reference the existing GCP config if already created for the request.
  if not request.connectCluster.gcpConfig:
    request.connectCluster.gcpConfig = {}
  request.connectCluster.gcpConfig.accessConfig = {"networkConfigs": []}
  # Subnets may not be provided during update
  if not args.primary_subnet:
    network_config = {"dns_domain_names": []}
  else:
    network_config = {
        "primary_subnet": args.primary_subnet,
        "additional_subnets": [],
        "dns_domain_names": [],
    }
  if not args.additional_subnet:
    args.additional_subnet = []
  network_config["additional_subnets"] = list(args.additional_subnet)
  if not args.dns_name:
    args.dns_name = []
  network_config["dns_domain_names"] = list(args.dns_name)
  request.connectCluster.gcpConfig.accessConfig.networkConfigs.append(
      encoding.DictToMessage(network_config, _MESSAGE.ConnectNetworkConfig)
  )
  if isinstance(
      # (if the request is an update)
      request,
      _MESSAGE.ManagedkafkaProjectsLocationsConnectClustersPatchRequest,
  ):
    request.updateMask = re.sub(
        r"gcpConfig\.accessConfig\.networkConfigs\.dnsDomainNames",
        "gcpConfig.accessConfig.networkConfigs",
        request.updateMask,
    )
    request.updateMask = re.sub(
        r"gcpConfig\.accessConfig\.networkConfigs\.primarySubnet",
        "gcpConfig.accessConfig.networkConfigs",
        request.updateMask,
    )
    request.updateMask = re.sub(
        r"gcpConfig\.accessConfig\.networkConfigs\.additionalSubnets",
        "gcpConfig.accessConfig.networkConfigs",
        request.updateMask,
    )
  return request


def PrepareConnectClusterCreate(_, args, request):
  """Load the config JSON from the argument to the request and build the kafka cluster resource path.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
  """
  if args.config_file:
    config = core.yaml.load(args.config_file)
    request.connectCluster.config = encoding.DictToMessage(
        config, _MESSAGE.ConnectCluster.ConfigValue
    )
  project = args.project or core.properties.VALUES.core.project.Get()
  # If the user provides the full path, we don't need to build it.
  kafka_cluster_path = args.kafka_cluster
  if not re.match(r"projects/.+/locations/.+/clusters/.+", args.kafka_cluster):
    location = args.location or args.connect_cluster.split("/")[3]
    kafka_cluster_path = (
        f"projects/{project}/locations/{location}/clusters/{args.kafka_cluster}"
    )
  request.connectCluster.kafkaCluster = kafka_cluster_path
  return request


def PrepareConnectClusterUpdate(_, args, request):
  """Map the update flags to the request and update mask.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
  """
  if args.config_file:
    config = core.yaml.load(args.config_file)
    request.connectCluster.config = encoding.DictToMessage(
        config, _MESSAGE.ConnectCluster.ConfigValue
    )
    request.updateMask = AppendUpdateMask(request.updateMask, "config")
  if args.clear_configs:
    request.updateMask = AppendUpdateMask(request.updateMask, "config")
  if args.clear_dns_names:
    request.updateMask = AppendUpdateMask(
        request.updateMask,
        "gcpConfig.accessConfig.networkConfigs.dnsDomainNames",
    )
  if args.clear_secrets:
    request.updateMask = AppendUpdateMask(
        request.updateMask, "gcpConfig.secretPaths"
    )
  if args.clear_labels:
    request.updateMask = AppendUpdateMask(request.updateMask, "labels")
  return request


def ConnectorCreateReadConfigAndTaskRestartPolicy(_, args, request):
  """Load the config JSON from the argument to the request.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
  """
  if not request.connector:
    request.connector = {}
  if args.config_file:
    config = core.yaml.load(args.config_file)
    request.connector.configs = encoding.DictToMessage(
        config, _MESSAGE.Connector.ConfigsValue
    )
  # Gcloud's duration formatting doesn't seem to play nice with the protobuf
  # duration parsing, so we convert to seconds and append the unit suffix here.
  task_restart_policy_dict = {}
  if args.task_restart_max_backoff:
    task_restart_policy_dict["maximum_backoff"] = (
        str(args.task_restart_max_backoff.total_seconds) + "s"
    )
  if args.task_restart_min_backoff:
    task_restart_policy_dict["minimum_backoff"] = (
        str(args.task_restart_min_backoff.total_seconds) + "s"
    )
  if args.task_retry_disabled:
    task_restart_policy_dict["task_retry_disabled"] = (
        str(args.task_retry_disabled)
    )
  if task_restart_policy_dict:
    request.connector.taskRestartPolicy = encoding.DictToMessage(
        task_restart_policy_dict,
        _MESSAGE.TaskRetryPolicy,
    )
  return request


def ConnectorUpdateReadConfigAndTaskRestartPolicy(_, args, request):
  """Load the config JSON from the argument to the request, and parse out the task restart policy.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
  """
  if not request.connector:
    request.connector = {}
  if args.config_file:
    config = core.yaml.load(args.config_file)
    request.connector.configs = encoding.DictToMessage(
        config, _MESSAGE.Connector.ConfigsValue
    )
    request.updateMask = AppendUpdateMask(request.updateMask, "configs")
  # Gcloud's duration formatting doesn't seem to play nice with the protobuf
  # duration parsing, so we convert to seconds and append the unit suffix here.
  task_restart_policy_dict = {}
  if args.task_restart_max_backoff:
    task_restart_policy_dict["maximum_backoff"] = (
        str(args.task_restart_max_backoff.total_seconds) + "s"
    )
  if args.task_restart_min_backoff:
    task_restart_policy_dict["minimum_backoff"] = (
        str(args.task_restart_min_backoff.total_seconds) + "s"
    )
  if args.task_retry_disabled:
    task_restart_policy_dict["task_retry_disabled"] = (
        str(args.task_retry_disabled)
    )
  if task_restart_policy_dict:
    request.connector.taskRestartPolicy = encoding.DictToMessage(
        task_restart_policy_dict,
        _MESSAGE.TaskRetryPolicy,
    )
  return request


def PatchConfigs(_, args, request):
  """Unnest the configs dictionary to the update mask.

  Args:
    _:  resource parameter required but unused variable.
    args: list of flags.
    request:  the payload to return.

  Returns:
    The new update mask with the configs.
  """
  if args.configs:
    update_mask = request.updateMask.split(",")
    update_mask.remove("configs")
    configs_list = []
    for key, _ in args.configs.items():
      configs_list.append(f'configs["{key}"]')
    request.updateMask = AppendUpdateMask(
        ",".join(update_mask), ",".join(configs_list)
    )
  # This flag is guarded with a mutex so it won't conflict with the above.
  if args.clear_configs:
    request.updateMask = AppendUpdateMask(request.updateMask, "configs")
  return request


def ParseMode(mode) -> str:
  """Parse the mode enum to a string.

  Args:
    mode: The mode enum of the schema registry or subject.

  Returns:
    The mode string.
  """

  if mode == _MESSAGE.SchemaMode.ModeValueValuesEnum.READWRITE:
    return "READWRITE"
  elif mode == _MESSAGE.SchemaMode.ModeValueValuesEnum.READONLY:
    return "READONLY"
  elif mode == _MESSAGE.SchemaMode.ModeValueValuesEnum.IMPORT:
    return "IMPORT"
  else:
    return "NONE"


def ParseCompatibility(compatibility) -> str:
  """Parse the compatibility enum to a string.

  Args:
    compatibility: The compatibility enum of the schema registry or subject.

  Returns:
    The compatibility string.
  """
  if (
      compatibility
      == _MESSAGE.SchemaConfig.CompatibilityValueValuesEnum.BACKWARD
  ):
    return "BACKWARD"
  elif (
      compatibility
      == _MESSAGE.SchemaConfig.CompatibilityValueValuesEnum.BACKWARD_TRANSITIVE
  ):
    return "BACKWARD_TRANSITIVE"
  elif (
      compatibility
      == _MESSAGE.SchemaConfig.CompatibilityValueValuesEnum.FORWARD
  ):
    return "FORWARD"
  elif (
      compatibility
      == _MESSAGE.SchemaConfig.CompatibilityValueValuesEnum.FORWARD_TRANSITIVE
  ):
    return "FORWARD_TRANSITIVE"
  elif compatibility == _MESSAGE.SchemaConfig.CompatibilityValueValuesEnum.FULL:
    return "FULL"
  elif (
      compatibility
      == _MESSAGE.SchemaConfig.CompatibilityValueValuesEnum.FULL_TRANSITIVE
  ):
    return "FULL_TRANSITIVE"
  else:
    return "NONE"


def ParseProject(project_id=None):
  return project_id or properties.VALUES.core.project.Get(required=True)


def DeleteSubjectMode(subject, subject_run_resource, context):
  """Called when the user runs gcloud managed-kafka schema-registries subject delete ...

  Args:
    subject: The subject of the attribute to delete.
    subject_run_resource: The subject resource path.
    context: The context of the schema registry if provided.

  Returns:
    The updated subject with its mode deleted.
  """

  message = apis.GetMessagesModule("managedkafka", "v1")
  client = apis.GetClientInstance("managedkafka", "v1")

  schema_registry_resource = subject_run_resource

  schema_registry_resource = (
      f"{schema_registry_resource}{SUBJECTS_MODE_RESOURCE_PATH}{subject}"
  )

  # Check if context is provided.
  if context:
    log.status.Print("Deleting subject mode for [%s]." % subject)
    request = message.ManagedkafkaProjectsLocationsSchemaRegistriesContextsModeDeleteRequest(
        name=schema_registry_resource
    )
    try:
      client.projects_locations_schemaRegistries_contexts_mode.Delete(
          request=request
      )
      log.UpdatedResource(subject, details="mode. It is now unset.")
    except apitools_exceptions.HttpNotFoundError as e:
      api_error = exceptions.HttpException(e, HTTP_ERROR_FORMAT)
      log.status.Print(api_error.message)
      if "Resource not found" in api_error.message:
        raise exceptions.HttpException(
            e, error_format="Subject {} not found.".format(subject)
        )
  else:
    log.status.Print("Deleting subject mode for [%s]." % subject)
    request = (
        message.ManagedkafkaProjectsLocationsSchemaRegistriesModeDeleteRequest(
            name=schema_registry_resource
        )
    )
    try:
      client.projects_locations_schemaRegistries_mode.Delete(request=request)
      log.UpdatedResource(subject, details="mode. It is now unset.")
    except apitools_exceptions.HttpNotFoundError as e:
      api_error = exceptions.HttpException(e, HTTP_ERROR_FORMAT)
      log.status.Print(api_error.message)
      if "Resource not found" in api_error.message:
        raise exceptions.HttpException(
            e, error_format="Subject {} not found.".format(subject)
        )


def DeleteSubjectConfig(subject, schema_registry_resource, context):
  """Called when the user runs gcloud managed-kafka schema-registries subject delete ...

  Args:
    subject: The subject of the attribute to delete.
    schema_registry_resource: The schema registry resource path.
    context: The context of the schema registry if provided.

  Returns:
    The updated subject with its config deleted.
  """
  message = apis.GetMessagesModule("managedkafka", "v1")
  client = apis.GetClientInstance("managedkafka", "v1")
  name = f"{schema_registry_resource}{SUBJECTS_CONFIG_RESOURCE_PATH}{subject}"
  # Check if context is provided.
  if context:
    log.status.Print("Deleting subject config for [%s]." % subject)
    request = message.ManagedkafkaProjectsLocationsSchemaRegistriesContextsConfigDeleteRequest(
        name=name
    )
    try:
      client.projects_locations_schemaRegistries_contexts_config.Delete(
          request=request
      )
      log.UpdatedResource(subject, details="config. It is now unset.")
    except apitools_exceptions.HttpNotFoundError as e:
      api_error = exceptions.HttpException(e, HTTP_ERROR_FORMAT)
      log.status.Print(api_error.message)
      if "Resource not found" in api_error.message:
        raise exceptions.HttpException(
            e, error_format="Subject {} not found.".format(subject)
        )
  else:
    log.status.Print("Deleting subject config for [%s]." % subject)
    request = message.ManagedkafkaProjectsLocationsSchemaRegistriesConfigDeleteRequest(
        name=name
    )
    try:
      client.projects_locations_schemaRegistries_config.Delete(request=request)
      log.UpdatedResource(subject, details="config. It is now unset.")
    except apitools_exceptions.HttpNotFoundError as e:
      api_error = exceptions.HttpException(e, HTTP_ERROR_FORMAT)
      log.status.Print(api_error.message)
      if "Resource not found" in api_error.message:
        raise exceptions.HttpException(
            e, error_format="Subject {} not found.".format(subject)
        )


AdditionalSubnetDeprecationBeforeAppendAction = actions.DeprecationAction(
    flag_name="--additional-subnet",
    warn="The --additional-subnet flag is deprecated and will be removed in a "
    "future version. Managed Kafka Connect clusters can now reach any endpoint "
    "accessible from the primary subnet without the need to define additional "
    "subnets. Please see "
    "https://cloud.google.com/managed-service-for-apache-kafka/docs/connect-cluster/create-connect-cluster#worker-subnet "
    "for more information.",
    removed=False,
    action="append")

HTTP_ERROR_FORMAT = (
    "ResponseError: code={status_code}, message={status_message}"
)



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for exporting an App Engine image to Artifact Registry."""

from googlecloudsdk.api_lib.app import appengine_api_client
from googlecloudsdk.core import exceptions


def export_image(
    project: str,
    service: str,
    version: str,
    destination_repository: str,
    api_client: appengine_api_client.AppengineApiClient,
) -> appengine_api_client.ExportImageResult:
  """Exports the App Engine image to Artifact Registry.

  Args:
    project: The project ID.
    service: The service ID.
    version: The version ID.
    destination_repository: The destination Artifact Registry repository.
    api_client: The App Engine API client.

  Returns:
    An ExportImageResult object.

  Raises:
    exceptions.Error: If image export fails.
  """

  try:
    return api_client.ExportImageAndWait(
        app_id=project,
        service_id=service,
        version_id=version,
        destination_repository=destination_repository,
    )
  except exceptions.Error as e:
    raise exceptions.Error(f'Failed to export image: {e!r}') from e
